"""This module contains interactive plots for Pastas models."""

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from plotly.subplots import make_subplots
from scipy.stats import norm, probplot

from pastas.extensions import register_model_accessor
from pastas.plotting.plotutil import (
    _get_height_ratios,
    _table_formatter_params,
    _table_formatter_stderr,
)
from pastas.rfunc import HantushWellModel
from pastas.stats import acf


@register_model_accessor("plotly")
class Plotly:
    """Extension class for interactive plotly figures for pastas Models.

    Examples
    --------
    >>> ps.extensions.register_plotly_extension()
    INFO: Registered plotly plotting methods in Model class, e.g. `ml.plotly.plot()`.
    >>> fig = ml.plotly.results()
    >>> fig.write_html("results_figure.html")

    Methods
    -------
    plot
        plot oseries and model simulation, interactive version of `ml.plot()`
    results
        plot oseries, model simulation, contribution, step responses and parameters
        table,interactive version of `ml.plots.results()`
    diagnostics
        plot noise, autocorrelation, distribution of noise and heteroscedasticity,
        interactive version of `ml.plots.diagnostics()`
    """

    def __init__(self, model):
        self._model = model

    def plot(self, tmin=None, tmax=None):
        """Plotly version of pastas.Model.plot().

        Parameters
        ----------
        tmin : pd.Timestamp, optional
            start time for model simulation, by default None
        tmax : pd.Timestamp, optional
            end time for model simulation, by default None

        Returns
        -------
        fig : plotly.Figure
            plotly Figure showing oseries and model simulation
        """

        traces = []

        o = self._model.observations()
        o_nu = self._model.oseries.series.drop(o.index)

        # add oseries
        if not o_nu.empty:
            trace_oseries_nu = go.Scattergl(
                x=o_nu.index,
                y=o_nu.values,
                mode="markers",
                marker={"color": "gray", "size": 3},
                name="(unused)",
                legendgroup="oseries",
                showlegend=False,
            )
            trace_oseries = go.Scattergl(
                x=o.index,
                y=o.values,
                mode="markers",
                marker={"color": "black", "size": 5},
                name=self._model.oseries.name,
                legendgroup="oseries",
            )
            traces.append(trace_oseries_nu)
            traces.append(trace_oseries)
        else:
            trace_oseries = go.Scattergl(
                x=o.index,
                y=o.values,
                mode="markers",
                marker_color="black",
                name=self._model.oseries.name,
                legendgroup="oseries",
            )
            traces.append(trace_oseries)

        sim = self._model.simulate(tmin=tmin, tmax=tmax)
        trace_sim = go.Scattergl(
            x=sim.index,
            y=sim.values,
            mode="lines",
            marker_color="#1F77B4",
            name=f"Sim (R<sup>2</sup> = {self._model.stats.rsq():.3f})",
            legendgroup="sim",
        )
        traces.append(trace_sim)

        layout = {
            "xaxis": {"range": [sim.index[0], sim.index[-1]]},
            "yaxis": {"title": "(m NAP)"},
            "legend": {
                "traceorder": "reversed+grouped",
                "orientation": "h",
                "xanchor": "left",
                "yanchor": "bottom",
                "x": 0.0,
                "y": 1.02,
            },
            "dragmode": "pan",
            "margin": dict(t=70, b=40, l=40, r=10),
        }

        return go.Figure(data=traces, layout=go.Layout(layout))

    def results(self, tmin=None, tmax=None, stderr=False):
        """Plotly version of pastas.Model.plots.results().

        Parameters
        ----------
        ml : pastas.Model
            model to plot results for
        tmin : pd.Timestamp, optional
            start time for model results, by default None
        tmax : pd.Timestamp, optional
            end time for model results, by default None
        stderr : bool, optional
            include standard errors in parameter table, by default False

        Returns
        -------
        dict
            dictionary containing plotly subplots data and layout
        """
        if tmin is None:
            tmin = self._model.settings["tmin"]
        if tmax is None:
            tmax = self._model.settings["tmax"]

        # for collecting data
        traces = []

        # dimensions
        nsm = len(self._model.stressmodels)
        nrows = 2 + nsm
        naxes = 2 + 2 * nsm

        # oseries
        o = self._model.observations(tmin=tmin, tmax=tmax)
        o_nu = self._model.oseries.series.drop(o.index)

        trace_oseries_nu = go.Scattergl(
            x=o_nu.index,
            y=o_nu.values,
            mode="markers",
            marker={"color": "gray", "size": 3},
            name="(unused)",
            legendgroup="0",
            showlegend=False,
            xaxis="x",
            yaxis="y",
        )
        traces.append(trace_oseries_nu)

        trace_oseries = go.Scattergl(
            x=o.index,
            y=o.values,
            mode="markers",
            marker={"color": "black", "size": 5},
            name=self._model.oseries.name,
            legendgroup="1",
            xaxis="x",
            yaxis="y",
        )
        traces.append(trace_oseries)

        # simulation
        sim = self._model.simulate(tmin=tmin, tmax=tmax)
        trace_sim = go.Scattergl(
            x=sim.index,
            y=sim.values,
            mode="lines",
            marker_color="#1F77B4",
            name=f"simulation (R<sup>2</sup> = {self._model.stats.rsq():.3f})",
            legendgroup="2",
            xaxis="x",
            yaxis="y",
        )
        traces.append(trace_sim)

        # residuals
        res = self._model.residuals(tmin=tmin, tmax=tmax)
        trace_res = go.Scattergl(
            x=res.index,
            y=res.values,
            mode="lines",
            marker_color="black",
            name="residuals",
            # legendgroup="residuals",
            xaxis="x",
            yaxis="y2",
            showlegend=False,
        )
        traces.append(trace_res)

        # noise
        if self._model.settings["noise"]:
            noise = self._model.noise(tmin=tmin, tmax=tmax)
            trace_noise = go.Scattergl(
                x=noise.index,
                y=noise.values,
                mode="lines",
                marker_color="#1F77B4",
                name="noise",
                # legendgroup="residuals",
                xaxis="x",
                yaxis="y2",
                showlegend=False,
            )
            traces.append(trace_noise)

        # contributions
        contribs = self._model.get_contributions(
            split=False,
            tmin=tmin,
            tmax=tmax,
            # return_warmup=return_warmup,
        )
        for i, c in enumerate(contribs):
            iax_contrib = 3 + i
            iax_response = 3 + nsm + i

            trace_c = go.Scattergl(
                x=c.index,
                y=c.values,
                mode="lines",
                marker_color="#1F77B4",
                name=f"{c.name}",
                legendgroup=f"contrib{i}",
                xaxis="x",
                yaxis=f"y{iax_contrib}",
                showlegend=False,
            )
            traces.append(trace_c)

            # response
            rkwargs = {}
            p = None
            if self._model.stressmodels[c.name].rfunc is not None:
                if isinstance(self._model.stressmodels[c.name].rfunc, HantushWellModel):
                    rkwargs = {"warn": False}
                    p = self._model.stressmodels[c.name].get_parameters(
                        model=self._model, istress=0
                    )
            response = self._model._get_response(
                block_or_step="step", name=c.name, p=p, add_0=True, **rkwargs
            )
            trace_r = go.Scatter(
                x=response.index,
                y=response.values,
                mode="lines",
                marker_color="#1F77B4",
                name=f"{c.name}",
                # legendgroup=f"contrib{i}",
                xaxis=f"x{3 + nsm}",
                yaxis=f"y{iax_response}",
                showlegend=False,
            )
            traces.append(trace_r)

        # calculate subplot
        ylims = [
            (
                min([sim.min(), o[tmin:tmax].min()]),
                max([sim.max(), o[tmin:tmax].max()]),
            ),
            (res.min(), res.max()),
        ]  # use residuals (b/c always bigger than noise)

        for contrib in contribs:
            hs = contrib.loc[tmin:tmax]
            if hs.empty:
                if contrib.empty:
                    ylims.append((0.0, 0.0))
                else:
                    ylims.append((contrib.min(), hs.max()))
            else:
                ylims.append((hs.min(), hs.max()))
        hrs = np.asarray(_get_height_ratios(ylims))
        # claim at least 1% figure height for zero contributions for visibility
        hrs = hrs.clip(min=0.01 * np.sum(hrs))
        # subplot positions and spacing
        dx = 0.015
        dy = 0.01
        # calculate whitespace
        wspace = (2 * len(hrs) - 2) * dy
        # scale plot heights
        hrs_frac = hrs / np.sum(hrs) * (1 - wspace)
        # add whitespace
        hrs_frac[[0, -1]] += dy
        hrs_frac[1:-1] += 2 * dy
        # calculate y-mid positions (between plots)
        y_pos = 1 - np.cumsum(np.concatenate([np.zeros(1), hrs_frac]))
        x_pos = 0.67
        # get tops and bottoms of axes from y-position array
        ytops = y_pos[:-1].copy()
        ytops[1:] -= dy  # half of whitespace by lowering tops by dy
        ybots = y_pos[1:].copy()
        ybots[:-1] += dy  # create other half of whitespace by raising bottoms by dy
        ybots[ybots < 0] = 0  # floating point issues, should always be > 0

        # parameter table
        col_names = ["name", "optimal", "stderr"] if stderr else ["name", "optimal"]
        p = self._model.parameters.copy().loc[:, col_names]
        p.loc[:, "name"] = p.index
        optimal = p.loc[:, "optimal"].copy()
        p["optimal"] = p["optimal"].astype(str)
        p.loc[:, "optimal"] = optimal.apply(_table_formatter_params)

        values = ["<b>Parameter</b>", "<b>Optimal</b>"]
        if stderr:
            values.append("<b>Std. Err.</b>")
            stderr_values = p.loc[:, "stderr"] / optimal
            p["stderr"] = p["stderr"].astype(str)
            p.loc[:, "stderr"] = stderr_values.abs().apply(_table_formatter_stderr)

        tab = go.Table(
            domain=dict(x=[x_pos + dx, 1.0], y=[y_pos[2] + dy, 1.0]),
            header=dict(
                values=values,
                font=dict(size=12),
                align=["left", "center", "center"],
                height=40,
            ),
            cells=dict(
                values=[p[k].tolist() for k in p.columns],
                align=["left", "right", "right"],
                height=30,
            ),
            columnwidth=[100, 70, 70] if stderr else [100, 70],
        )

        traces.append(tab)

        layout_dict = dict(
            # top left (row 0, col 0) - oseries/simulation plot
            xaxis=dict(
                domain=[0.0, x_pos - dx],
                anchor=f"y{nrows}",
            ),
            yaxis=dict(
                domain=[ybots[0], ytops[0]],
                anchor="x",
            ),
            # row 1, col 0 - residuals/noise plot
            xaxis2=dict(
                domain=[0.0, x_pos - dx],
                anchor="y2",
            ),
            yaxis2=dict(
                domain=[ybots[1], ytops[1]],
                anchor="x2",
                scaleanchor="y",
            ),
        )
        # add layout for stressmodels
        # unfortunately I can only get it to work with numbering down the
        # columns, e.g.
        # ax1 table
        # ax2 table
        # ax3 ax5
        # ax4 ax6
        # also order is important, so first all contributions, then step responses

        # add contributions axes
        for ism in range(nsm):
            irow = 2 + ism  # 2, 3, 4, ..., (0-based)
            iax_contrib = 3 + ism  # 3, 4, 5, ..., (1-based)

            # contributions subplot
            layout_dict[f"xaxis{iax_contrib}"] = dict(
                domain=[0.0, x_pos - dx],
                anchor=f"y{iax_contrib}",
            )
            layout_dict[f"yaxis{iax_contrib}"] = dict(
                domain=[ybots[irow], ytops[irow]],
                anchor=f"x{iax_contrib}",
                scaleanchor="y",
            )
        # add response axes
        for ism in range(nsm):
            irow = 2 + ism  # 2, 3, 4, ..., (0-based)
            iax_response = 3 + nsm + ism

            # step response subplot
            layout_dict[f"xaxis{iax_response}"] = dict(
                domain=[x_pos + dx, 1],
                anchor=f"y{naxes}",
            )
            layout_dict[f"yaxis{iax_response}"] = dict(
                domain=[ybots[irow], ytops[irow]],
                anchor=f"x{3 + nsm}",
            )

        layout = go.Layout(layout_dict)

        fig = go.Figure(data=traces, layout=layout)

        # add titles for subplots
        rnlabel = [
            "residuals / noise" if self._model.settings["noise"] else "residuals"
        ]
        labels = rnlabel + list(self._model.stressmodels.keys())
        for i, lbl in enumerate(labels):
            fig.add_annotation(
                xref="paper",
                yref="paper",
                x=0.015,
                y=y_pos[i + 1],
                xanchor="left",
                yanchor="middle",
                showarrow=False,
                text=lbl,
                font=dict(
                    size=12,
                ),
                align="left",
            )

        # add table title
        fig.add_annotation(
            xref="paper",
            yref="paper",
            x=x_pos + dx + 0.015,
            y=1.015,
            xanchor="left",
            yanchor="middle",
            showarrow=False,
            text="Model parameters (n<sub>c</sub>=7)",
            font=dict(
                size=12,
            ),
            align="left",
        )

        # add step response title
        fig.add_annotation(
            xref="paper",
            yref="paper",
            x=x_pos + dx + 0.015,
            y=y_pos[2],
            xanchor="left",
            yanchor="middle",
            showarrow=False,
            text="Step response",
            font=dict(
                size=12,
            ),
            align="left",
        )

        # set size, axes labels and legend position
        fig.update_layout(
            xaxis={"range": [tmin, tmax]},
            yaxis={"title": "[m NAP]"},  # oseries, simulation
            yaxis2={"title": "[m]"},  # residuals
            legend={
                "traceorder": "reversed+grouped",
                "orientation": "h",
                "xanchor": "left",
                "yanchor": "bottom",
                "x": 0.0,
                "y": 1.00,
            },
            # height=1000,
            # width=1000,
            margin=dict(t=60, b=20, l=10, r=25),
            dragmode="pan",
        )
        update_labels = {f"yaxis{i}": {"title": "[m]"} for i in range(3, nrows + 1)}
        update_labels[f"xaxis{nrows + 1}"] = {"title": "time [d]"}
        fig.update_layout(**update_labels)

        return fig

    def diagnostics(self):
        """Plotly version of pastas.Model.plots.diagnostics().

        Parameters
        ----------
        ml : pastas.Model
            model to plot results for

        Returns
        -------
        fig : plotly.Figure
            plotly figure with model diagnostics
        """
        # prepare data
        sim = self._model.simulate()
        if self._model.settings["noise"]:
            series = self._model.noise()
            resnoisename = "noise"
        else:
            series = self._model.residuals()
            resnoisename = "residuals"

        df_acf = acf(series, full_output=True)
        x = df_acf.index.total_seconds() / (24 * 60 * 60)
        conf = df_acf["conf"].rolling(10, min_periods=1).mean().values

        if self._model.interpolate_simulation:
            sim_interpolated = np.interp(series.index.asi8, sim.index.asi8, sim.values)
            sim = pd.Series(index=series.index, data=sim_interpolated)

        sim = sim.loc[series.index]

        # let the plotting begin
        fig = make_subplots(
            rows=4,
            cols=2,
            specs=[
                [{"colspan": 2}, None],
                [{"colspan": 2}, None],
                [{}, {}],
                [{}, {}],
            ],
            subplot_titles=[
                f"{resnoisename} (&#956;={series.mean():.1f})",
                "Autocorrelation",
                "Histogram",
                "Probability plot",
                "Heteroscedasticity",
                "",
            ],
            horizontal_spacing=0.1,
            vertical_spacing=0.075,
        )

        # Noise
        fig.add_trace(
            go.Scatter(
                x=series.index,
                y=series.values,
                line_color="rgba(31,119,180,1)",
                showlegend=False,
                name=resnoisename,
            ),
            row=1,
            col=1,
        )

        # ACF
        fig.add_trace(
            go.Scatter(
                x=x,
                y=conf,
                mode="lines",
                fillcolor="rgba(32,146,230,0.3)",
                line_color="rgba(255,255,255,0)",
                showlegend=False,
                name="CI",
                hoverinfo="skip",
            ),
            row=2,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                x=x,
                y=-conf,
                mode="lines",
                fillcolor="rgba(32,146,230,0.3)",
                fill="tonexty",
                line_color="rgba(255,255,255,0)",
                legendgroup="1",
                showlegend=False,
                name="CI",
                hoverinfo="skip",
            ),
            row=2,
            col=1,
        )
        fig.add_trace(
            go.Bar(
                x=x,
                y=df_acf.iloc[:, 0],
                marker={"color": ["black"] * df_acf.index.size},
                legendgroup="0",
                showlegend=False,
                width=0.9,
                name="ACF",
            ),
            row=2,
            col=1,
        )

        # normality
        _, bins = np.histogram(series.values, bins=50)
        bins = 0.5 * (bins[:-1] + bins[1:])
        fig.add_trace(
            go.Histogram(
                x=series.values,
                ybins={"size": 50},
                name="Histogram",
                showlegend=False,
                histnorm="probability density",
                marker=dict(color="rgba(31,119,180,1)"),
            ),
            row=3,
            col=1,
        )
        pdf = norm.pdf(bins, series.mean(), series.std())
        fig.add_trace(
            go.Scatter(
                x=bins,
                y=pdf,
                mode="lines",
                name="PDF",
                line=dict(
                    dash="dash",
                    color="black",
                ),
                showlegend=False,
            ),
            row=3,
            col=1,
        )

        (osm, osr), (slope, intercept, r) = probplot(series, dist="norm", rvalue=False)
        fig.add_trace(
            go.Scatter(
                x=osm,
                y=osr,
                mode="markers",
                marker={"color": "rgba(31,119,180,1)"},
                showlegend=False,
                name="probplot",
            ),
            row=3,
            col=2,
        )

        fig.add_trace(
            go.Scatter(
                x=osm,
                y=slope * osm + intercept,
                mode="lines",
                line={"color": "black"},
                showlegend=False,
                name="fit",
            ),
            row=3,
            col=2,
        )
        fig.add_annotation(
            xref="x4 domain",
            yref="y4 domain",
            x=0.05,
            y=0.95,
            xanchor="left",
            yanchor="top",
            showarrow=False,
            text=f"R<sup>2</sup>={r:.3f}",
            font=dict(
                size=12,
            ),
            align="left",
        )

        # heteroscedasticity
        fig.add_trace(
            go.Scatter(
                name="heteroscedasticity",
                x=sim,
                y=series,
                mode="markers",
                showlegend=False,
                marker=dict(
                    size=5,
                    color="rgba(31,119,180,0.5)",
                    line=dict(
                        color="black",
                        width=0.25,
                    ),
                ),
            ),
            row=4,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                name="heteroscedasticity",
                x=sim,
                y=np.sqrt(series.abs()),
                mode="markers",
                showlegend=False,
                marker=dict(
                    size=5,
                    color="rgba(31,119,180,0.5)",
                    line=dict(
                        color="black",
                        width=0.25,
                    ),
                ),
            ),
            row=4,
            col=2,
        )

        # update x-axes, y-axes
        fig.update_yaxes(title_text=resnoisename, title_standoff=0, row=1, col=1)

        fig.update_xaxes(title_text="Lags [d]", title_standoff=0, row=2, col=1)
        fig.update_yaxes(title_text="ACF [-]", title_standoff=0, row=2, col=1)

        fig.update_yaxes(
            title_text="Probability density", title_standoff=0, row=3, col=1
        )

        fig.update_xaxes(
            title_text="Theoretical quantiles", title_standoff=0, row=3, col=2
        )
        fig.update_yaxes(title_text="Ordered values", title_standoff=0, row=3, col=2)

        fig.update_xaxes(title_text="Simulated values", title_standoff=0, row=4, col=1)
        fig.update_yaxes(title_text="Residuals", title_standoff=0, row=4, col=1)

        fig.update_yaxes(
            title_text=r"&#8730; Residuals", title_standoff=0, row=4, col=2
        )
        fig.update_xaxes(title_text="Simulated values", title_standoff=0, row=4, col=2)

        # update titles

        fig.layout.annotations[0].update(x=0.05, font={"size": 13})
        fig.layout.annotations[1].update(x=0.05, font={"size": 13})
        fig.layout.annotations[2].update(x=0.05, font={"size": 13})
        fig.layout.annotations[3].update(x=0.60, font={"size": 13})
        fig.layout.annotations[4].update(x=0.05, font={"size": 13})

        #
        fig.update_layout(
            legend={
                "traceorder": "grouped",
                "orientation": "h",
                "xanchor": "left",
                "yanchor": "bottom",
                "x": 0.0,
                "y": 1.00,
            },
            # width=750,
            # height=1000,
            margin=dict(t=60, b=20, l=25, r=10),
        )
        return fig



from typing import Any
from unittest.mock import patch

from pastas.version import check_numba_scipy, get_versions, show_versions


def test_check_numba_scipy_not_installed() -> None:
    with patch("importlib.import_module", side_effect=ImportError):
        assert check_numba_scipy() is False


def test_check_numba_scipy_version_mismatch() -> None:
    with (
        patch("importlib.import_module") as mock_import_module,
        patch("importlib.metadata.version") as mock_version,
        patch("importlib.metadata.requires") as mock_requires,
    ):
        mock_import_module.return_value = None
        mock_version.side_effect = lambda x: "1.6.0" if x == "scipy" else None
        mock_requires.return_value = ["scipy<=1.5.0"]

        assert check_numba_scipy() is False


def test_get_versions_basic() -> None:
    get_versions()


def test_get_versions_optional() -> None:
    get_versions(optional=True)


def test_show_versions(capsys: Any) -> None:
    show_versions()



"""This module contains methods to compute the Standardized Groundwater Index."""

from numpy import array, linspace
from pandas import DataFrame, Series
from scipy.stats import norm


def sgi(series: Series, timescale_months: int = 1) -> Series:
    """Method to compute the Standardized Groundwater Index (SGI)
    :cite:t:`bloomfield_analysis_2013`.

    Parameters
    ----------
    series: pandas.Series or Pandas.DataFrame
        Pandas time series of the groundwater levels
        for which the SGI is to be determined
    timescale_months: integer, optional
        Length of the aggregation period in months (default: 1; allowed: 1, 2, 3)

    Returns
    -------
    sgi_series: pandas.Series or Pandas.DataFrame
        Pandas time series of the groundwater levels. Time series index should be a
        pandas DatetimeIndex.

    Notes
    -----
    The Standardized Groundwater Index (SGI) is a non-parametric method to
    standardize groundwater levels. The SGI is calculated for each aggregation
    period within the year separately. The data within that period in all years
    in the series are used to determine the reference for which the index is
    calculated for each value in that period.
    The SGI is a dimensionless index and is used to compare groundwater levels
    across different wells. It may be useful to resample the time series to a
    monthly interval before computing the SGI.
    """
    if timescale_months not in (1, 2, 3):
        raise ValueError(
            "SGI can only be called with timescale_months = 1, 2, or 3; not"
            + str(timescale_months)
        )
    if isinstance(series, DataFrame):
        series = series.apply(sgi, timescale_months=timescale_months)
    elif isinstance(series, Series):
        # Create a copy to ensure series is untouched.
        # Set dtype to avoid conflict when assigning SGI values
        series = series.copy().dropna().astype(float)

        # Loop over the months
        for month in range(1, 13, timescale_months):
            sel = array(range(timescale_months)) + month
            data = series[series.index.month.isin(sel)]
            n = data.size  # Number of observations
            pmin = 1 / (2 * n)
            pmax = 1 - pmin
            sgi_values = norm.ppf(linspace(pmin, pmax, n))
            series.loc[data.sort_values().index] = sgi_values

    return series



"""This module contains all the stress models available in Pastas.

Stress models are used to translate an input time series into contribution that
explains (part of) the output series.

Examples
--------

>>> sm = ps.StressModel(stress, rfunc=ps.Gamma(), name="sm1")
>>> ml.add_stressmodel(stressmodel=sm)

See Also
--------
pastas.model.Model.add_stressmodel
"""

from inspect import isclass
from logging import getLogger
from typing import Any

import numpy as np
from packaging.version import parse as parse_version
from pandas import DataFrame, Series, Timedelta, Timestamp, concat, date_range
from pandas import __version__ as pd_version
from scipy.signal import fftconvolve

from pastas.typing import (
    ArrayLike,
    Model,
    Recharge,
    RFunc,
    StressSettingsDict,
    TimestampType,
)

from .decorators import njit, set_parameter
from .recharge import Linear
from .rfunc import Exponential, HantushWellModel, One
from .timeseries import TimeSeries
from .utils import validate_name

pandas_version = parse_version(pd_version)

logger = getLogger(__name__)

__all__ = [
    "StressModel",
    "Constant",
    "StepModel",
    "LinearTrend",
    "RechargeModel",
    "WellModel",
    "TarsoModel",
    "ChangeModel",
]


class StressModelBase:
    """StressModel Base class called by each StressModel object.

    Attributes
    ----------
    name: str
        Name of this stressmodel object. Used as prefix for the parameters.
    parameters: pandas.DataFrame
        The DataFrame containing the parameters.
    """

    _name = "StressModelBase"

    def __init__(
        self,
        name: str,
        tmin: TimestampType,
        tmax: TimestampType,
        rfunc: RFunc | None = None,
        up: bool = True,
        gain_scale_factor: float = 1.0,
    ) -> None:
        self.name = validate_name(name)
        self.tmin = tmin
        self.tmax = tmax
        self.freq = None

        if rfunc is not None:
            if isclass(rfunc):
                raise TypeError(
                    "the rfunc argument must be an instance of response function, not "
                    "a class. Please provide an instance, e.g., ps.Exponential()"
                )
            rfunc.update_rfunc_settings(up=up, gain_scale_factor=gain_scale_factor)
        self.rfunc = rfunc

        self.parameters = DataFrame(
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"]
        )

        self.stress = []

    @property
    def nparam(self) -> tuple[int]:
        return self.parameters.index.size

    def set_init_parameters(self) -> None:
        """Set the initial parameters (back) to their default values."""

    @set_parameter
    def _set_initial(self, name: str, value: float) -> None:
        """Internal method to set the initial parameter value.

        Notes
        -----
        The preferred method for parameter setting is through the model.
        """
        self.parameters.at[name, "initial"] = value

    @set_parameter
    def _set_pmin(self, name: str, value: float) -> None:
        """Internal method to set the lower bound of the parameter value.

        Notes
        -----
        The preferred method for parameter setting is through the model.
        """
        self.parameters.at[name, "pmin"] = value

    @set_parameter
    def _set_pmax(self, name: str, value: float) -> None:
        """Internal method to set the upper bound of the parameter value.

        Notes
        -----
        The preferred method for parameter setting is through the model.
        """
        self.parameters.at[name, "pmax"] = value

    @set_parameter
    def _set_vary(self, name: str, value: float) -> None:
        """Internal method to set if the parameter is varied during optimization.

        Notes
        -----
        The preferred method for parameter setting is through the model.
        """
        self.parameters.at[name, "vary"] = bool(value)

    @set_parameter
    def _set_dist(self, name: str, value: str) -> None:
        """Internal method to set distribution of prior of the parameter.

        Notes
        -----
        The preferred method for parameter setting is through the model.
        """
        self.parameters.at[name, "dist"] = str(value)

    def update_stress(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
    ) -> None:
        """Method to update the settings of the all stresses in the stress model.

        Parameters
        ----------
        freq: str, optional
            String representing the desired frequency of the time series. Must be one
            of the following: (D, h, m, s, ms, us, ns) or a multiple of that e.g. "7D".
        tmin: str or pandas.Timestamp, optional
            String that can be converted to, or a Pandas Timestamp with the minimum
            time of the series.
        tmax: str or pandas.Timestamp, optional
            String that can be converted to, or a Pandas Timestamp with the maximum
            time of the series.

        Notes
        -----
        For the individual options for the different settings please refer to the
        docstring from the TimeSeries.update_series() method.

        See Also
        --------
        ps.timeseries.TimeSeries.update_series
        """
        for stress in self.stress:
            stress.update_series(freq=freq, tmin=tmin, tmax=tmax)

        if freq:
            self.freq = freq

    def get_stress(
        self,
        p: ArrayLike | None = None,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        istress: int | None = None,
        **kwargs,
    ) -> DataFrame:
        """Returns the stress(es) of the time series object as a pandas DataFrame.

        If the time series object has multiple stresses each column represents a stress.

        Returns
        -------
        stress: pandas.Dataframe
            Pandas dataframe of the stress(es)
        """
        if tmin is None:
            tmin = self.tmin
        if tmax is None:
            tmax = self.tmax

        self.update_stress(tmin=tmin, tmax=tmax, freq=freq)

        return self.stress[0].series

    def to_dict(self, **kwargs):
        """Method to export the stress model object."""

    def get_nsplit(self) -> int:
        """Determine in how many time series the contribution can be split."""
        if hasattr(self, "nsplit"):
            return self.nsplit
        else:
            return len(self.stress)

    def _get_block(
        self, p: ArrayLike, dt: float, tmin: TimestampType, tmax: TimestampType
    ) -> ArrayLike:
        """Internal method to get the block-response function."""
        if tmin is not None and tmax is not None:
            day = Timedelta(1, "D")
            maxtmax = (Timestamp(tmax) - Timestamp(tmin)) / day
        else:
            maxtmax = None
        b = self.rfunc.block(p, dt, maxtmax=maxtmax)
        return b

    def get_settings(self) -> dict:
        """Method to obtain the settings of the stresses.

        Returns
        -------
        settings: dict

        Notes
        -----
        To update the settings of the time series, use the `update_stress` method.
        """
        if len(self.stress) == 0:
            settings = None
        else:
            settings = {stress.name: stress.settings for stress in self.stress}
        return settings

    def get_parameters(self, model=None) -> ArrayLike:
        """Get parameters and return as array.

        Parameters
        ----------
        model : pastas.Model, optional
            If provided, and the model is solved, return optimal model parameter-values.
            Otherwise, return initial parameter-values.
        istress : int, optional
            if provided, return specific parameter set, else return all parameters.

        Returns
        -------
        p : array_like
            An array of the parameters of the stressmodel.
        """
        if model is None:
            p = self.parameters.initial.values
        else:
            p = model.get_parameters(self.name)
        return p


class StressModel(StressModelBase):
    """Stress model convoluting a stress with a response function.

    Parameters
    ----------
    stress: pandas.Series
        pandas.Series with pandas.DatetimeIndex containing the stress.
    rfunc: pastas.rfunc instance
        An instance of the response function used in the convolution with the stress.
    name: str
        Name of the stress.
    up: bool or None, optional
        True if response function is positive (default), False if negative. None if
        you don't want to define if response is positive or negative.
    settings: dict or str, optional
        The settings of the stress. This can be a string referring to a predefined
        settings dictionary (defined in ps.rcParams["timeseries"]), or a dictionary with
        the settings to apply. For more information refer to Time series settings
        section below.
    metadata: dict, optional
        dictionary containing metadata about the stress. This is passed onto the
        TimeSeries object.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.

    Other Parameters
    ----------------

    Time series settings

    fill_nan : {"drop", "mean", "interpolate"} or float
        Method for filling NaNs.
           * `drop`: drop NaNs from time series
           * `mean`: fill NaNs with mean value of time series
           * `interpolate`: fill NaNs by interpolating between finite values
           * `float`: fill NaN with provided value, e.g. 0.0
    fill_before : {"mean", "bfill"} or float
        Method for extending time series into past.
           * `mean`: extend time series into past with mean value of time series
           * `bfill`: extend time series into past by back-filling first value
           * `float`: extend time series into past with provided value, e.g. 0.0
    fill_after : {"mean", "ffill"} or float
        Method for extending time series into future.
           * `mean`: extend time series into future with mean value of time series
           * `ffill`: extend time series into future by forward-filling last value
           * `float`: extend time series into future with provided value, e.g. 0.0
    sample_up : {"mean", "interpolate", "divide"} or float
        Method for up-sampling time series (increasing frequency, e.g. going from weekly
        to daily values).
           * `bfill` or `backfill`: fill up-sampled time steps by back-filling current
             values
           * `ffill` or `pad`: fill up-sampled time steps by forward-filling current
             values
           * `mean`: fill up-sampled time steps with mean of timeseries
           * `interpolate`: fill up-sampled time steps by interpolating between current
             values
           * `divide`: fill up-sampled steps with current value divided by length of
             current time steps (i.e. spread value over new time steps).
    sample_down : {"mean", "drop", "sum", "min", "max"}
        Method for down-sampling time series (decreasing frequency, e.g. going from
        daily to weekly values).
           * `mean`: resample time series by taking the mean
           * `drop`: resample the time series by taking the mean, dropping any
             NaN-values
           * `sum`: resample time series by summing values
           * `max`: resample time series with maximum value
           * `min`: resample time series with minimum value

    Examples
    --------
    >>> import pastas as ps
    >>> import pandas as pd
    >>> sm = ps.StressModel(stress=pd.Series(), rfunc=ps.Gamma(), name="Prec",
    >>>                     settings="prec")

    See Also
    --------
    pastas.rfunc
    pastas.timeseries.TimeSeries
    """

    _name = "StressModel"

    def __init__(
        self,
        stress: Series,
        rfunc: RFunc,
        name: str,
        up: bool = True,
        settings: str | StressSettingsDict | None = None,
        metadata: dict | None = None,
        gain_scale_factor: float | None = None,
    ) -> None:
        stress = TimeSeries(stress, settings=settings, metadata=metadata)

        StressModelBase.__init__(
            self,
            name=name,
            tmin=stress.series.index.min(),
            tmax=stress.series.index.max(),
            rfunc=rfunc,
            up=up,
            gain_scale_factor=(
                stress.series.std() if gain_scale_factor is None else gain_scale_factor
            ),
        )

        self.gain_scale_factor = gain_scale_factor
        self.freq = stress.settings["freq"]
        self.stress.append(stress)
        self.set_init_parameters()

    def set_init_parameters(self) -> None:
        """Set the initial parameters (back) to their default values."""
        self.parameters = self.rfunc.get_init_parameters(self.name)

    def simulate(
        self,
        p: ArrayLike,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        dt: float = 1.0,
    ) -> Series:
        """Simulates the head contribution.

        Parameters
        ----------
        p: array_like
            array_like object with the values as floats representing the model
            parameters.
        tmin: str, optional
        tmax: str, optional
        freq: str, optional
        dt: int, optional

        Returns
        -------
        pandas.Series
            The simulated head contribution.
        """
        self.update_stress(tmin=tmin, tmax=tmax, freq=freq)
        b = self._get_block(p, dt, tmin, tmax)
        stress = self.stress[0].series
        npoints = stress.index.size
        h = Series(
            data=fftconvolve(stress, b, "full")[:npoints],
            index=stress.index,
            name=self.name,
        )
        return h

    def to_dict(self, series: bool = True) -> dict:
        """Method to export the StressModel object.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the StressModel
            object.

        Notes
        -----
        Settings and metadata are exported with the stress.
        """
        data = {
            "class": self._name,
            "rfunc": self.rfunc.to_dict(),
            "name": self.name,
            "up": self.rfunc.up,
            "stress": self.stress[0].to_dict(series=series),
            "gain_scale_factor": self.gain_scale_factor,
        }
        return data


class StepModel(StressModelBase):
    """Stressmodel that simulates a step trend.

    Parameters
    ----------
    tstart: str or Timestamp
        String with the start date of the step, e.g. '2018-01-01'. This value is
        fixed by default. Use ml.set_parameter("step_tstart", vary=True) to vary the
        start time of the step trend.
    name: str
        String with the name of the stressmodel.
    rfunc: pastas.rfunc instance
        Pastas response function used to simulate the effect of the step. Default is
        ps.rfunc.One(), an instant effect.
    up: bool, optional
        Force a direction of the step. Default is None.

    Notes
    -----
    The step trend is calculated as follows. First, a binary series is created,
    with zero values before tstart, and ones after the start. This series is
    convolved with the block response to simulate a step trend.
    """

    _name = "StepModel"

    def __init__(
        self,
        tstart: TimestampType,
        name: str,
        rfunc: RFunc | None = None,
        up: bool = None,
    ) -> None:
        if rfunc is None:
            rfunc = One()
        StressModelBase.__init__(
            self,
            name=name,
            tmin=Timestamp.min,
            tmax=Timestamp.max,
            rfunc=rfunc,
            up=up,
        )
        self.tstart = Timestamp(tstart)
        self.set_init_parameters()

    def set_init_parameters(self) -> None:
        self.parameters = self.rfunc.get_init_parameters(self.name)
        tmin = Timestamp.min.toordinal()
        tmax = Timestamp.max.toordinal()
        tinit = self.tstart.toordinal()

        self.parameters.loc[self.name + "_tstart"] = (
            tinit,
            tmin,
            tmax,
            False,
            self.name,
            "uniform",
        )

    def simulate(
        self,
        p: ArrayLike,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        dt: float = 1.0,
    ) -> Series:
        tstart = Timestamp.fromordinal(int(p[-1]))
        tindex = date_range(tmin, tmax, freq=freq)
        h = Series(0, tindex, name=self.name)
        h.loc[h.index > tstart] = 1

        b = self._get_block(p[:-1], dt, tmin, tmax)
        npoints = h.index.size
        h = Series(
            data=fftconvolve(h, b, "full")[:npoints],
            index=h.index,
            name=self.name,
        )
        return h

    def to_dict(self, **kwargs) -> dict:
        """Method to export the StepModel object.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct object.
        """
        data = {
            "class": self._name,
            "tstart": self.tstart,
            "name": self.name,
            "rfunc": self.rfunc.to_dict(),
            "up": self.rfunc.up,
        }
        return data


class LinearTrend(StressModelBase):
    """Stressmodel that simulates a linear trend.

    Parameters
    ----------
    start: str
        String with a date to start the trend (e.g., "2018-01-01"), will be
        transformed to an ordinal number internally.
    end: str
        String with a date to end the trend (e.g., "2018-01-01"), will be transformed
        to an ordinal number internally.
    name: str, optional
        String with the name of the stress model.

    Notes
    -----
    While possible, it is not recommended to vary the parameters for the start and
    end time of the linear trend. These parameters are usually hard or even impossible
    to estimate from the data.
    """

    _name = "LinearTrend"

    def __init__(
        self, start: TimestampType, end: TimestampType, name: str = "trend"
    ) -> None:
        StressModelBase.__init__(
            self, name=name, tmin=Timestamp.min, tmax=Timestamp.max
        )
        self.start = start
        self.end = end
        self.set_init_parameters()

    def set_init_parameters(self) -> None:
        """Set the initial parameters for the stress model."""
        start = Timestamp(self.start).toordinal()
        end = Timestamp(self.end).toordinal()
        tmin = Timestamp.min.toordinal()
        tmax = Timestamp.max.toordinal()

        self.parameters.loc[self.name + "_a"] = (
            0.0,
            -np.inf,
            np.inf,
            True,
            self.name,
            "uniform",
        )
        self.parameters.loc[self.name + "_tstart"] = (
            start,
            tmin,
            tmax,
            False,
            self.name,
            "uniform",
        )
        self.parameters.loc[self.name + "_tend"] = (
            end,
            tmin,
            tmax,
            False,
            self.name,
            "uniform",
        )

    def simulate(
        self,
        p: ArrayLike,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        dt: float = 1.0,
    ) -> Series:
        """Simulate the trend."""
        tindex = date_range(tmin, tmax, freq=freq)

        if p[1] < tindex[0].toordinal():
            tmin = tindex[0]
        else:
            tmin = Timestamp.fromordinal(int(p[1]))

        if p[2] >= tindex[-1].toordinal():
            tmax = tindex[-1]
        else:
            tmax = Timestamp.fromordinal(int(p[2]))

        trend = tindex.to_series().diff() / Timedelta(1, "D")
        trend.loc[:tmin] = 0
        trend.loc[tmax:] = 0
        trend = trend.cumsum() * p[0]
        return trend.rename(self.name)

    def to_dict(self, **kwargs) -> dict:
        """Method to export a dictionary to reconstruct the stressmodel.

        Parameters
        ----------
        kwargs

        Returns
        -------
        data: dict
        """
        data = {
            "class": self._name,
            "start": self.start,
            "end": self.end,
            "name": self.name,
        }
        return data


class Constant(StressModelBase):
    """A constant value that is added to the time series model.

    Parameters
    ----------
    name: str, optional
        Name of the stressmodel.
    initial: float, optional
        Initial estimate of the parameter value. For example, the minimum of the
        observed series.
    """

    _name = "Constant"

    def __init__(self, name: str = "constant", initial: float = 0.0) -> None:
        StressModelBase.__init__(
            self, name=name, tmin=Timestamp.min, tmax=Timestamp.max
        )
        self.initial = initial
        self.set_init_parameters()

    def set_init_parameters(self):
        self.parameters.loc[self.name + "_d"] = (
            self.initial,
            np.nan,
            np.nan,
            True,
            self.name,
            "uniform",
        )

    @staticmethod
    def simulate(p: float | None = None) -> float:
        return p

    def to_dict(self, **kwargs):
        """Method to export the StressModel object.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the StressModel
            object.
        """
        data = {
            "class": self._name,
            "name": self.name,
            "initial": self.initial,
        }
        return data


class WellModel(StressModelBase):
    """Convolution of one or more stresses with a single scaled response function.

    Parameters
    ----------
    stress: list
        list containing the stresses time series.
    name: str
        name of the stressmodel.
    distances: array_like
        array_like of distances between the stresses (wells) and the oseries
        (monitoring well), must be in the same order as the stresses. This
        distance is used to scale the HantushWellModel response function for
        each stress.
    rfunc: pastas.rfunc instance, optional
        this model only works with the HantushWellModel response function, default is
        None which will initialize a HantushWellModel response function.
    up: bool, optional
        whether a positive stress has an increasing or decreasing effect on the model,
        by default False, in which case positive stress lowers e.g., the groundwater
        level.
    settings: str, list of dict, optional
        The settings of the stress. By default this is "well". This can be a string
        referring to a predefined settings dictionary (defined in
        ps.rcParams["timeseries"]), or a dictionary with the settings to apply. For more
        information, refer to Time series settings section below.
    sort_wells: bool, optional
        sort wells from closest to furthest, by default True.

    Other Parameters
    ----------------

    Time series settings

    fill_nan : {"drop", "mean", "interpolate"} or float
        Method for filling NaNs.
           * `drop`: drop NaNs from time series
           * `mean`: fill NaNs with mean value of time series
           * `interpolate`: fill NaNs by interpolating between finite values
           * `float`: fill NaN with provided value, e.g. 0.0
    fill_before : {"mean", "bfill"} or float
        Method for extending time series into past.
           * `mean`: extend time series into past with mean value of time series
           * `bfill`: extend time series into past by back-filling first value
           * `float`: extend time series into past with provided value, e.g. 0.0
    fill_after : {"mean", "ffill"} or float
        Method for extending time series into future.
           * `mean`: extend time series into future with mean value of time series
           * `ffill`: extend time series into future by forward-filling last value
           * `float`: extend time series into future with provided value, e.g. 0.0
    sample_up : {"mean", "interpolate", "divide"} or float
        Method for up-sampling time series (increasing frequency, e.g. going from weekly
        to daily values).
           * `bfill` or `backfill`: fill up-sampled time steps by back-filling current
             values
           * `ffill` or `pad`: fill up-sampled time steps by forward-filling current
             values
           * `mean`: fill up-sampled time steps with mean of timeseries
           * `interpolate`: fill up-sampled time steps by interpolating between current
             values
           * `divide`: fill up-sampled steps with current value divided by length of
             current time steps (i.e. spread value over new time steps).
    sample_down : {"mean", "drop", "sum", "min", "max"}
        Method for down-sampling time series (decreasing frequency, e.g. going from
        daily to weekly values).
           * `mean`: resample time series by taking the mean
           * `drop`: resample the time series by taking the mean, dropping any
             NaN-values
           * `sum`: resample time series by summing values
           * `max`: resample time series with maximum value
           * `min`: resample time series with minimum value

    Notes
    -----
    This class implements convolution of multiple series with the same response
    function. This can be applied when dealing with multiple wells in a time series
    model. The distance(s) from the pumping well(s) to the monitoring well have to be
    provided for each stress. See :cite:t:`brakenhoff_application_2022` for more details on the methods for this model.

    Only works with the HantushWellModel response function.
    """

    _name = "WellModel"

    def __init__(
        self,
        stress: list[Series],
        name: str,
        distances: ArrayLike,
        rfunc: RFunc | None = None,
        up: bool = False,
        settings: str | StressSettingsDict = "well",
        sort_wells: bool = True,
        metadata: list[dict[str, Any]] = None,
    ) -> None:
        # check response function
        if rfunc is None:
            rfunc = HantushWellModel()
        elif not isinstance(rfunc, HantushWellModel):
            raise NotImplementedError(
                "WellModel only supports the rfunc HantushWellModel!"
            )

        # check if number of stresses and distances match
        if len(stress) != len(distances):
            msg = (
                "The number of stresses does not match the number of distances "
                "provided."
            )
            logger.error(msg)
            raise ValueError(msg)
        else:
            self.distances = Series(
                index=[s.squeeze().name for s in stress],
                data=distances,
                name="distances",
            )

        # parse settings input
        if settings is None or isinstance(settings, str) or isinstance(settings, dict):
            settings = len(stress) * [settings]

        # if metadata is passed as dict -> convert to list
        if metadata is not None and isinstance(metadata, dict):
            metadata = [metadata]

        # parse stresses input
        stress = self._handle_stress(stress, settings, metadata)

        # sort wells by distance
        self.sort_wells = sort_wells
        if self.sort_wells:
            stress = [
                s for _, s in sorted(zip(distances, stress), key=lambda pair: pair[0])
            ]
            self.distances.sort_values(inplace=True)

        # estimate gain_scale_factor w/ max of stresses stdev
        gain_scale_factor = np.max([s.series.std() for s in stress])

        tmin = np.min([s.series.index.min() for s in stress])
        tmax = np.max([s.series.index.max() for s in stress])

        StressModelBase.__init__(
            self,
            name=name,
            tmin=tmin,
            tmax=tmax,
            rfunc=rfunc,
            up=up,
            gain_scale_factor=gain_scale_factor,
        )

        self.rfunc.set_distances(self.distances.values)

        self.stress = stress
        self.freq = self.stress[0].settings["freq"]
        self.set_init_parameters()

    def set_init_parameters(self) -> None:
        self.parameters = self.rfunc.get_init_parameters(self.name)

    def simulate(
        self,
        p: ArrayLike | None = None,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        dt: float = 1.0,
        istress: int | None = None,
        **kwargs,
    ) -> Series:
        distances = self.get_distances(istress=istress)
        stress_df = self.get_stress(
            p=p, tmin=tmin, tmax=tmax, freq=freq, istress=istress, squeeze=False
        )
        h = Series(data=0, index=self.stress[0].series.index, name=self.name)
        for name, r in distances.items():
            stress = stress_df.loc[:, name]
            npoints = stress.index.size
            p_with_r = np.concatenate([p, np.array([r])])
            b = self._get_block(p_with_r, dt, tmin, tmax)
            c = fftconvolve(stress, b, "full")[:npoints]
            h = h.add(Series(c, index=stress.index), fill_value=0.0)
        if istress is not None:
            if isinstance(istress, list):
                h.name = self.name + "_" + "+".join(str(i) for i in istress)
            elif self.stress[istress].name is not None:
                h.name = self.stress[istress].name
            else:
                h.name = self.name + "_" + str(istress)
        else:
            h.name = self.name
        return h

    @staticmethod
    def _handle_stress(stress, settings, metadata):
        """Internal method to handle user provided stress in init.

        Parameters
        ----------
        stress: pandas.Series, list or dict
            stress or collection of stresses.
        settings: dict or iterable
            settings dictionary.
        metadata : dict or list of dict
            metadata dictionaries corresponding to stress

        Returns
        -------
        stress: list
            return a list with the stresses transformed to pastas TimeSeries.
        """
        data = []

        if isinstance(stress, Series):
            data.append(TimeSeries(stress, settings=settings, metadata=metadata))
        elif isinstance(stress, dict):
            for i, (name, value) in enumerate(stress.items()):
                if metadata is not None:
                    imeta = metadata[i]
                else:
                    imeta = None
                data.append(
                    TimeSeries(value, name=name, settings=settings[i], metadata=imeta)
                )
        elif isinstance(stress, list):
            for i, value in enumerate(stress):
                if metadata is not None:
                    imeta = metadata[i]
                else:
                    imeta = None
                data.append(TimeSeries(value, settings=settings[i], metadata=imeta))
        else:
            msg = "Cannot parse 'stress' input. Provide a Series, dict or list."
            logger.error(msg)
            raise TypeError(msg)

        return data

    def get_stress(
        self,
        p: ArrayLike | None = None,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        istress: int | None = None,
        squeeze: bool = True,
        **kwargs,
    ) -> DataFrame:
        if tmin is None:
            tmin = self.tmin
        if tmax is None:
            tmax = self.tmax

        self.update_stress(tmin=tmin, tmax=tmax, freq=freq)

        if istress is None:
            df = DataFrame.from_dict({s.name: s.series for s in self.stress})
            if squeeze:
                return df.squeeze()
            else:
                return df
        elif isinstance(istress, list):
            return DataFrame.from_dict({s.name: s.series for s in self.stress}).iloc[
                :, istress
            ]
        else:
            if squeeze:
                return self.stress[istress].series
            else:
                return self.stress[istress].series.to_frame()

    def get_distances(self, istress: int | None = None) -> DataFrame:
        if istress is None:
            return self.distances
        elif isinstance(istress, list):
            return self.distances.iloc[istress]
        else:
            return self.distances.iloc[istress : istress + 1]

    def get_parameters(self, model=None, istress: int | None = None) -> ArrayLike:
        """Get parameters including distance to observation point and return as array
        (dimensions = (nstresses, 4)).

        Parameters
        ----------
        model : pastas.Model, optional
            If provided, and the model is solved, return optimal model parameter-values.
            Otherwise, return initial parameter-values.
        istress : int, optional
            if provided, return specific parameter set, else return all parameters.

        Returns
        -------
        p : array_like
            parameters for each stress as row of array, if istress is used returns
            only one row.
        """
        if model is None:
            p = self.parameters.initial.values
        else:
            p = model.get_parameters(self.name)

        distances = self.get_distances(istress=istress).values
        if distances.size > 1:
            p_with_r = np.concatenate(
                [np.tile(p, (distances.size, 1)), distances[:, np.newaxis]], axis=1
            )
        else:
            p_with_r = np.r_[p, distances]
        return p_with_r

    def dump_stress(self, series: bool = True) -> list:
        """Method to dump all stresses in the stresses list.

        Parameters
        ----------
        series: bool, optional
            True if time series are to be exported, False if only the name
            of the time series are needed. Settings are always exported.

        Returns
        -------
        data: dict
            dictionary with the dump of the stresses.
        """
        data = []

        for stress in self.stress:
            stress.name = validate_name(stress.name, raise_error=True)
            data.append(stress.to_dict(series=series))

        return data

    def to_dict(self, series: bool = True) -> dict:
        """Method to export the WellModel object.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the WellModel
            object.
        """
        data = {
            "class": self._name,
            "stress": self.dump_stress(series),
            "rfunc": self.rfunc.to_dict(),
            "name": self.name,
            "distances": self.distances.to_list(),
            "up": True if self.rfunc.up else False,
            "sort_wells": self.sort_wells,
        }
        return data

    def variance_gain(
        self, model: Model, istress: int | None = None, r: ArrayLike | None = None
    ) -> float:
        """Calculate variance of the gain for WellModel.

        Variance of the gain is calculated based on propagation of uncertainty using
        optimal parameter values and the estimated variances of A and b and the
        covariance between A and b.

        Parameters
        ----------
        model : pastas.Model
            optimized model
        istress : int or list of int, optional
            index of stress(es) for which to calculate variance of gain
        r : array_like, optional
            radial distance(s) at which to calculate variance of the gain,
            only considered if istress is None

        Returns
        -------
        var_gain : float
            variance of the gain calculated from model results for parameters A and b.

        See Also
        --------
        pastas.HantushWellModel.variance_gain
        """
        if model.solver is None:
            raise AttributeError("Model not optimized! Run solve() first!")
        if self.rfunc._name != "HantushWellModel":
            raise ValueError("Response function must be HantushWellModel!")
        if model.solver.pcov.isna().all(axis=None):
            model.logger.warning("Covariance matrix contains only NaNs!")

        # get parameters and (co)variances
        A = model.parameters.at[self.name + "_A", "optimal"]
        b = model.parameters.at[self.name + "_b", "optimal"]
        var_A = model.solver.pcov.at[self.name + "_A", self.name + "_A"]
        var_b = model.solver.pcov.at[self.name + "_b", self.name + "_b"]
        cov_Ab = model.solver.pcov.at[self.name + "_A", self.name + "_b"]

        if istress is None and r is None:
            r = np.asarray(self.distances)
        elif isinstance(istress, int) or isinstance(istress, list):
            if r is not None:
                logger.warning("kwarg 'r' is only used if istress is None!")
            r = self.distances.iloc[istress]
        elif istress is not None and r is None:
            raise ValueError("Parameter 'istress' must be None, list or int!")

        return self.rfunc.variance_gain(A, b, var_A, var_b, cov_Ab, r=r)


class RechargeModel(StressModelBase):
    """Stressmodel simulating the effect of groundwater recharge on the head.

    Parameters
    ----------
    prec: pandas.Series
        pandas.Series with pandas.DatetimeIndex containing the precipitation series.
        The precipitation series should be provided in mm/day when a nonlinear model is
        used.
    evap: pandas.Series
        pandas.Series with pandas.DatetimeIndex containing the potential evaporation
        series. The evaporation series should be provided in mm/day when a nonlinear
        model is used.
    rfunc: pastas.rfunc instance, optional
        Instance of the response function used in the convolution with the stress.
        Default is ps.Exponential().
    name: str, optional
        Name of the stress. Default is "recharge".
    recharge: pastas.recharge instance, optional
        Instance of a recharge model. Options are: Linear, FlexModel and Berendrecht.
        These can be accessed through ps.rch. Default is ps.rch.Linear().
    temp: pandas.Series, optional
        pandas.Series with pandas.DatetimeIndex containing the temperature series.
        It depends on the recharge model if this argument is required or not. The
        temperature series should be provided in degrees Celsius.
    settings: list of dicts or str, optional
        The settings of the precipitation, evaporation and optionally temperature time
        series, in this order. By default ("prec", "evap", "evap"). This can be a string
        referring to a predefined settings dict (defined in ps.rcParams["timeseries"]),
        or a dict with the settings to apply. For more information refer to Time Series
        Settings section below for more information.
    metadata: tuple of dicts or list of dicts, optional
        dictionary containing metadata about the stress. This is passed onto the
        TimeSeries object.

    Examples
    --------
    >>> sm = ps.RechargeModel(rain, evap, rfunc=ps.Exponential(),
    >>>                       recharge=ps.rch.FlexModel(), name="rch")
    >>> ml.add_stressmodel(sm)

    Other Parameters
    ----------------

    Time series settings

    fill_nan : {"drop", "mean", "interpolate"} or float
        Method for filling NaNs.
           * `drop`: drop NaNs from time series
           * `mean`: fill NaNs with mean value of time series
           * `interpolate`: fill NaNs by interpolating between finite values
           * `float`: fill NaN with provided value, e.g. 0.0
    fill_before : {"mean", "bfill"} or float
        Method for extending time series into past.
           * `mean`: extend time series into past with mean value of time series
           * `bfill`: extend time series into past by back-filling first value
           * `float`: extend time series into past with provided value, e.g. 0.0
    fill_after : {"mean", "ffill"} or float
        Method for extending time series into future.
           * `mean`: extend time series into future with mean value of time series
           * `ffill`: extend time series into future by forward-filling last value
           * `float`: extend time series into future with provided value, e.g. 0.0
    sample_up : {"mean", "interpolate", "divide"} or float
        Method for up-sampling time series (increasing frequency, e.g. going from weekly
        to daily values).
           * `bfill` or `backfill`: fill up-sampled time steps by back-filling current
             values
           * `ffill` or `pad`: fill up-sampled time steps by forward-filling current
             values
           * `mean`: fill up-sampled time steps with mean of timeseries
           * `interpolate`: fill up-sampled time steps by interpolating between current
             values
           * `divide`: fill up-sampled steps with current value divided by length of
             current time steps (i.e. spread value over new time steps).
    sample_down : {"mean", "drop", "sum", "min", "max"}
        Method for down-sampling time series (decreasing frequency, e.g. going from
        daily to weekly values).
           * `mean`: resample time series by taking the mean
           * `drop`: resample the time series by taking the mean, dropping any
             NaN-values
           * `sum`: resample time series by summing values
           * `max`: resample time series with maximum value
           * `min`: resample time series with minimum value

    Notes
    -----
    This stress model computes the contribution of precipitation and potential
    evaporation in two steps. In the first step a recharge flux is computed by a
    model determined by the input argument `recharge`. In the second step this
    recharge flux is convolved with a response function to obtain the contribution
    of recharge to the groundwater levels. If a nonlinear recharge model is used, the
    precipitation should be in mm/d.

    Warnings
    --------
    We recommend not to store a RechargeModel is a variable named `rm`. This name is
    already reserved in IPython to remove files and will cause problems later.

    Raises
    ------
    A warning if the the maximum annual precipitation is smaller than 12 and a
    nonlinear recharge model is applied. This is likely an indication that the units of
    the precipitation series are in m/d instead of mm/d. Please check the units of the
    precipitation series.

    See Also
    --------
    pastas.rfunc
    pastas.timeseries.TimeSeries
    pastas.recharge
    """

    _name = "RechargeModel"

    def __init__(
        self,
        prec: Series,
        evap: Series,
        rfunc: RFunc | None = None,
        name: str = "recharge",
        recharge: Recharge | None = None,
        temp: Series | None = None,
        settings: tuple[
            str | StressSettingsDict,
            str | StressSettingsDict,
            str | StressSettingsDict,
        ] = (
            "prec",
            "evap",
            "evap",
        ),
        metadata: tuple[dict | None, dict | None, dict | None] = (None, None, None),
    ) -> None:
        if rfunc is None:
            rfunc = Exponential()

        if recharge is None:
            recharge = Linear()

        # Store the precipitation and evaporation time series
        self.prec = TimeSeries(prec, settings=settings[0], metadata=metadata[0])
        self.evap = TimeSeries(evap, settings=settings[1], metadata=metadata[1])

        # Store recharge object
        self.recharge = recharge

        # Store a temperature time series if provided/needed or set to None
        if self.recharge.snow is True and temp is None:
            msg = (
                "Recharge model requires a temperature series. No temperature series "
                "were provided."
            )
            raise TypeError(msg)
        if temp is not None:
            if len(settings) < 3 or len(metadata) < 3:
                msg = "Number of values for the settings and/or metadata is incorrect."
                raise TypeError(msg)
            else:
                self.temp = TimeSeries(temp, settings=settings[2], metadata=metadata[2])
        else:
            self.temp = None

        # Select indices from validated stress where both series are available.
        index = self.prec.series.index.intersection(self.evap.series.index)
        if index.empty:
            msg = (
                "The stresses that were provided have no overlapping time indices. "
                "Please make sure the indices of the time series overlap."
            )
            logger.error(msg)
            raise Exception(msg)

        # Calculate initial recharge estimation for initial rfunc parameters
        p = self.recharge.get_init_parameters().initial.values
        gain_scale_factor = self.get_stress(
            p=p, tmin=index.min(), tmax=index.max(), freq=self.prec.settings["freq"]
        ).std()

        StressModelBase.__init__(
            self,
            name=name,
            tmin=index.min(),
            tmax=index.max(),
            rfunc=rfunc,
            up=True,
            gain_scale_factor=gain_scale_factor,
        )

        self.stress = [self.prec, self.evap]
        if self.temp:
            self.stress.append(self.temp)
        self.freq = self.prec.settings["freq"]
        self.set_init_parameters()
        if isinstance(self.recharge, Linear):
            self.nsplit = 2
        else:
            self.nsplit = 1

            # Check if precipitation is likely in mm/d and not m/d. If the maximum
            # value of the annual sums is smaller than 12 (m/d), the highest annual
            # precipitation in the world, then the precipitation is very likely in m/d
            # and not in mm/d. In this case a warning is given for nonlinear models.

            freq_offset = "YE" if pandas_version >= parse_version("2.2.0") else "A"
            if self.prec.series.resample(freq_offset).sum().max() < 12:
                msg = (
                    "The maximum annual precipitation is smaller than 12 m/d. Please "
                    "double-check if the stresses are in mm/d and not in m/d."
                )
                logger.warning(msg)

    def set_init_parameters(self) -> None:
        """Internal method to set the initial parameters."""
        self.parameters = concat(
            [
                self.rfunc.get_init_parameters(self.name),
                self.recharge.get_init_parameters(self.name),
            ]
        )

    def update_stress(
        self,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
    ) -> None:
        """Method to update the settings of the all stresses in the stress model.

        Parameters
        ----------
        freq: str, optional
            String representing the desired frequency of the time series. Must be one
            of the following: (D, h, m, s, ms, us, ns) or a multiple of that e.g. "7D".
        tmin: str or pandas.Timestamp, optional
            String that can be converted to, or a Pandas Timestamp with the minimum
            time of the series.
        tmax: str or pandas.Timestamp, optional
            String that can be converted to, or a Pandas Timestamp with the maximum
            time of the series.

        Notes
        -----
        For the individual options for the different settings please refer to the
        docstring from the TimeSeries.update_series() method.

        See Also
        --------
        ps.timeseries.TimeSeries.update_series
        """
        self.prec.update_series(freq=freq, tmin=tmin, tmax=tmax)
        self.evap.update_series(freq=freq, tmin=tmin, tmax=tmax)
        if self.temp is not None:
            self.temp.update_series(freq=freq, tmin=tmin, tmax=tmax)

        if freq:
            self.freq = freq

    def simulate(
        self,
        p: ArrayLike | None = None,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        dt: float = 1.0,
        istress: int | None = None,
        **kwargs,
    ) -> Series:
        """Method to simulate the contribution of recharge to the head.

        Parameters
        ----------
        p: array_like, optional
            array_like object with the values as floats representing the model
            parameters.
        tmin: string, optional
        tmax: string, optional
        freq: string, optional
        dt: float, optional
            Time step to use in the recharge calculation.
        istress: int, optional
            This only works for the Linear model!

        Returns
        -------
        pandas.Series
        """
        if p is None:
            p = self.parameters.initial.values
        b = self._get_block(p[: self.rfunc.nparam], dt, tmin, tmax)
        stress = self.get_stress(
            p=p, tmin=tmin, tmax=tmax, freq=freq, istress=istress
        ).values
        name = self.name

        if istress is not None:
            if istress == 1 and self.nsplit > 1:
                # only happen when Linear is used as the recharge model
                stress = stress * p[-1]
            if self.stress[istress].name is not None:
                name = f"{self.name} ({self.stress[istress].name})"

        return Series(
            data=fftconvolve(stress, b, "full")[: stress.size],
            index=self.prec.series.index,
            name=name,
        )

    def get_stress(
        self,
        p: ArrayLike | None = None,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        istress: int | None = None,
        **kwargs,
    ) -> Series:
        """Method to obtain the recharge stress calculated by the model.

        Parameters
        ----------
        p: array_like, optional
            array_like object with the values as floats representing the model
            parameters.
        tmin: string, optional
        tmax: string, optional
        freq: string, optional
        istress: int, optional
            Return one of the stresses used for the recharge calculation. 0 for
            precipitation, 1 for evaporation and 2 for temperature.
        kwargs

        Returns
        -------
        stress: pandas.Series
            When no istress is selected, this return the estimated recharge flux that
            is convolved with a response function on the simulate method.
        """
        if tmin is None:
            tmin = self.tmin
        if tmax is None:
            tmax = self.tmax

        self.update_stress(tmin=tmin, tmax=tmax, freq=freq)

        if istress is None:
            prec = self.prec.series.values
            evap = self.evap.series.values
            temp = None
            if self.temp is not None:
                temp = self.temp.series.values
            if p is None:
                p = self.parameters.initial.values
            stress = self.recharge.simulate(
                prec=prec, evap=evap, p=p[-self.recharge.nparam :], **{"temp": temp}
            )
            return Series(
                data=stress,
                index=self.prec.series.index,
                name="recharge",
            )
        elif istress == 0:
            return self.prec.series
        elif istress == 1:
            return self.evap.series
        else:
            return self.temp.series

    def get_water_balance(
        self,
        p: ArrayLike | None = None,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
    ) -> DataFrame:
        """Method to obtain the water balance components.

        Parameters
        ----------
        p: array_like, optional
            array_like object with the values as floats representing the model
            parameters.
        tmin: string, optional
        tmax: string, optional
        freq: string, optional

        Returns
        -------
        wb: pandas.DataFrame
            Dataframe with the water balance components, both fluxes and states.

        Notes
        -----
        This method return a data frame with all water balance components, fluxes and
        states. All ingoing fluxes have a positive sign (e.g., precipitation) and all
        outgoing fluxes have negative sign (e.g., recharge).

        Warnings
        --------
        This is an experimental method and may change in the future.

        Examples
        --------
        >>> sm = ps.RechargeModel(prec, evap, ps.Gamma(), ps.rch.FlexModel(),
        >>>                       name="rch")
        >>> ml.add_stressmodel(sm)
        >>> ml.solve()
        >>> wb = sm.get_water_balance(ml.get_parameters("rch"))
        >>> wb.plot(subplots=True)
        """
        if p is None:
            p = self.parameters.initial.values

        prec = self.get_stress(tmin=tmin, tmax=tmax, freq=freq, istress=0).values
        evap = self.get_stress(tmin=tmin, tmax=tmax, freq=freq, istress=1).values

        if self.temp is not None:
            temp = self.get_stress(tmin=tmin, tmax=tmax, freq=freq, istress=2).values
        else:
            temp = None
        df = self.recharge.get_water_balance(
            prec=prec, evap=evap, temp=temp, p=p[-self.recharge.nparam :]
        )
        df.index = self.prec.series.index
        return df

    def get_parameters(self, model=None, istress: int | None = None) -> ArrayLike:
        """Get parameters and return as array.

        Parameters
        ----------
        model : pastas.Model, optional
            If provided, and the model is solved, return optimal model parameter-values.
            Otherwise, return initial parameter-values.
        istress : int, optional
            if provided, return specific parameter set, else return all parameters.

        Returns
        -------
        p : array_like
            An array of the parameters of the stressmodel.
        """
        if model is None:
            p = self.parameters.initial.values
        else:
            p = model.get_parameters(self.name)

        if istress is not None and isinstance(self.recharge, Linear):
            if istress == 0:
                p = p[:-1]
            elif istress == 1:
                p[0] *= p[-1]
                p = p[:-1]
        return p

    def to_dict(self, series: bool = True) -> dict:
        """Method to export the RechargeModel object.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the object.

        Notes
        -----
        Settings and metadata are exported with the stress.
        """
        data = {
            "class": self._name,
            "prec": self.prec.to_dict(series=series),
            "evap": self.evap.to_dict(series=series),
            "rfunc": self.rfunc.to_dict(),
            "name": self.name,
            "recharge": self.recharge.to_dict(),
            "temp": self.temp.to_dict() if self.temp else None,
        }
        return data


class TarsoModel(RechargeModel):
    """Stressmodel simulating the effect of recharge using the Tarso method.

    Parameters
    ----------
    prec: pandas.Series
        pandas.Series with pandas.DatetimeIndex containing the precipitation series.
    evap: pandas.Series
        pandas.Series with pandas.DatetimeIndex containing the potential evaporation
        series.
    oseries: pandas.Series, optional
        A pandas.Series with pandas.DatetimeIndex of observations to which the model
        will be calibrated. It is used to determine the initial values of the
        drainage levels and the boundaries of the upper drainage level. Specify
        either oseries or dmin and dmax.
    dmin: float, optional
        The minimum drainage level. It is used to determine the initial values of the
        drainage levels and the lower boundary of the upper drainage level. Specify
        either oseries or dmin and dmax.
    dmax : float, optional
        The maximum drainage level. It is used to determine the initial values of the
        drainage levels and the upper boundary of the upper drainage level. Specify
        either oseries or dmin and dmax.
    rfunc: pastas.rfunc instance
        this model only works with the Exponential response function.

    See Also
    --------
    pastas.recharge

    Notes
    -----
    The Threshold autoregressive self-exciting open-loop (Tarso) model
    :cite:t:`knotters_tarso_1999` is nonlinear in structure because it incorporates
    two regimes which are separated by a threshold. This model method can be used to
    simulate a groundwater system where the groundwater head reaches the surface or
    drainage level in wet conditions. TarsoModel uses two drainage levels, with two
    exponential response functions. When the simulation reaches the second drainage
    level, the second response function becomes active. Because of its structure,
    TarsoModel cannot be combined with other stress models, a constant or a transform.
    TarsoModel inherits from RechargeModel. Only parameters specific to the child
    class are named above.
    """

    _name = "TarsoModel"

    def __init__(
        self,
        prec: Series,
        evap: Series,
        oseries: Series | None = None,
        dmin: float | None = None,
        dmax: float | None = None,
        rfunc: RFunc | None = None,
        **kwargs,
    ) -> None:
        if oseries is not None:
            if dmin is not None or dmax is not None:
                msg = "Please specify either oseries or dmin and dmax"
                raise (Exception(msg))
            dmin = oseries.min()
            dmax = oseries.max()
        elif dmin is None or dmax is None:
            msg = "Please specify either oseries or dmin and dmax"
            raise (Exception(msg))
        if rfunc is None:
            rfunc = Exponential()
        if not isinstance(rfunc, Exponential):
            raise NotImplementedError("TarsoModel only supports rfunc Exponential!")
        self.dmin = dmin
        self.dmax = dmax
        super().__init__(prec=prec, evap=evap, rfunc=rfunc, **kwargs)
        self.nsplit = 1

    def set_init_parameters(self) -> None:
        # parameters for the first drainage level
        p0 = self.rfunc.get_init_parameters(self.name)
        initial = self.dmin + 0.5 * (self.dmax - self.dmin)
        pd0 = Series(
            {
                "initial": initial,
                "pmin": np.nan,
                "pmax": np.nan,
                "vary": True,
                "name": self.name,
                "dist": "uniform",
            }
        )
        p0.loc[f"{self.name}_d"] = pd0
        p0.index = [f"{x}0" for x in p0.index]

        # parameters for the second drainage level
        p1 = self.rfunc.get_init_parameters(self.name)
        initial = self.dmin + 0.75 * (self.dmax - self.dmin)
        pd1 = Series(
            {
                "initial": initial,
                "pmin": self.dmin,
                "pmax": self.dmax,
                "vary": True,
                "name": self.name,
                "dist": "uniform",
            }
        )
        p1.loc[f"{self.name}_d"] = pd1
        p1.index = [f"{x}1" for x in p1.index]

        # parameters for the recharge-method
        pr = self.recharge.get_init_parameters(self.name)

        # combine all parameters
        self.parameters = concat([p0, p1, pr])

    def simulate(
        self,
        p: ArrayLike | None = None,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq=None,
        dt: float = 1.0,
    ) -> Series:
        stress = self.get_stress(p=p, tmin=tmin, tmax=tmax, freq=freq)
        h = self.tarso(p[: -self.recharge.nparam], stress.values, dt)
        sim = Series(h, name=self.name, index=stress.index)
        return sim

    def to_dict(self, series: bool = True) -> dict:
        """Method to export the TarsoModel object.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the object.

        Notes
        -----
        Settings and metadata are exported with the stress.
        """
        data = super().to_dict(series)
        data["dmin"] = self.dmin
        data["dmax"] = self.dmax
        return data

    @staticmethod
    def _check_stressmodel_compatibility(ml: Model) -> None:
        """Internal method to check if no other stressmodels, a constants or a transform
        is used."""
        msg = (
            "A TarsoModel cannot be combined with %s. Either remove the TarsoModel or "
            "the %s."
        )
        if len(ml.stressmodels) > 1:
            logger.warning(msg, "other stressmodels", "stressmodels")
        if ml.constant is not None:
            logger.warning(msg, "a constant", "constant")
        if ml.transform is not None:
            logger.warning(msg, "a transform", "transform")

    @staticmethod
    @njit
    def tarso(p: ArrayLike, r: ArrayLike, dt: float) -> ArrayLike:
        """Calculates the head based on exponential decay of the previous timestep and
        recharge, using two thresholds."""
        A0, a0, d0, A1, a1, d1 = p

        # calculate physical meaning of these parameters
        S0 = a0 / A0
        c0 = A0

        S1 = a1 / A1
        c1 = A1

        # calculate effective parameters for the top level
        c_e = 1 / ((1 / c0) + (1 / c1))
        d_e = (c1 / (c0 + c1)) * d0 + (c0 / (c0 + c1)) * d1
        a_e = S1 * c_e

        h = np.full(len(r), np.nan)
        for i in range(len(r)):
            if i == 0:
                h0 = (d0 + d1) / 2
                high = h0 > d1
                if high:
                    S, a, c, d = S1, a_e, c_e, d_e
                else:
                    S, a, c, d = S0, a0, c0, d0
            else:
                h0 = h[i - 1]
            exp_a = np.exp(-dt / a)
            h[i] = (h0 - d) * exp_a + r[i] * c * (1 - exp_a) + d
            newhigh = h[i] > d1
            if high != newhigh:
                # calculate time until d1 is reached
                dtdr = -S * c * np.log((d1 - d - r[i] * c) / (h0 - d - r[i] * c))
                if dtdr > dt:
                    raise (Exception())
                # change parameters
                high = newhigh
                if high:
                    S, a, c, d = S1, a_e, c_e, d_e
                else:
                    S, a, c, d = S0, a0, c0, d0
                # calculate new level after reaching d1
                exp_a = np.exp(-(dt - dtdr) / a)
                h[i] = (d1 - d) * exp_a + r[i] * c * (1 - exp_a) + d
        return h


class ChangeModel(StressModelBase):
    """Model where the response function changes from one to another over time.

    Parameters
    ----------
    stress: pandas.Series
        pandas Series object containing the stress.
    rfunc1: pastas.rfunc instance
        The instance of the response function used in the convolution with the stress.
    rfunc2: pastas.rfunc instance
        The instance of the response function used in the convolution with the stress.
    name: str
        name of the stress.
    tchange: str
        string with the approximate date of the change.
    up: bool or None, optional
        True if response function is positive (default), False if negative. None if
        you don't want to define if response is positive or negative.
    settings: dict or str, optional
        The settings of the stress. This can be a string referring to a predefined
        settings dict (defined in ps.rcParams["timeseries"]), or a dict with the
        settings to apply. For more information, refer to the docs of pastas.Timeseries
        for further information.
    metadata: dict, optional
        dictionary containing metadata about the stress. This is passed onto the
        TimeSeries object.

    Other Parameters
    ----------------

    Time series settings

    fill_nan : {"drop", "mean", "interpolate"} or float
        Method for filling NaNs.
           * `drop`: drop NaNs from time series
           * `mean`: fill NaNs with mean value of time series
           * `interpolate`: fill NaNs by interpolating between finite values
           * `float`: fill NaN with provided value, e.g. 0.0
    fill_before : {"mean", "bfill"} or float
        Method for extending time series into past.
           * `mean`: extend time series into past with mean value of time series
           * `bfill`: extend time series into past by back-filling first value
           * `float`: extend time series into past with provided value, e.g. 0.0
    fill_after : {"mean", "ffill"} or float
        Method for extending time series into future.
           * `mean`: extend time series into future with mean value of time series
           * `ffill`: extend time series into future by forward-filling last value
           * `float`: extend time series into future with provided value, e.g. 0.0
    sample_up : {"mean", "interpolate", "divide"} or float
        Method for up-sampling time series (increasing frequency, e.g. going from weekly
        to daily values).
           * `bfill` or `backfill`: fill up-sampled time steps by back-filling current
             values
           * `ffill` or `pad`: fill up-sampled time steps by forward-filling current
             values
           * `mean`: fill up-sampled time steps with mean of timeseries
           * `interpolate`: fill up-sampled time steps by interpolating between current
             values
           * `divide`: fill up-sampled steps with current value divided by length of
             current time steps (i.e. spread value over new time steps).
    sample_down : {"mean", "drop", "sum", "min", "max"}
        Method for down-sampling time series (decreasing frequency, e.g. going from
        daily to weekly values).
           * `mean`: resample time series by taking the mean
           * `drop`: resample the time series by taking the mean, dropping any
             NaN-values
           * `sum`: resample time series by summing values
           * `max`: resample time series with maximum value
           * `min`: resample time series with minimum value

    Notes
    -----
    This model is based on :cite:t:`obergfell_identification_2019`.
    """

    _name = "ChangeModel"

    def __init__(
        self,
        stress: Series,
        rfunc1: RFunc,
        rfunc2: RFunc,
        name: str,
        tchange: str | TimestampType,
        up: bool = True,
        settings: str | StressSettingsDict | None = None,
        metadata: dict | None = None,
    ) -> None:
        stress = TimeSeries(stress, settings=settings, metadata=metadata)

        StressModelBase.__init__(
            self,
            name=name,
            rfunc=None,
            tmin=stress.series.index.min(),
            tmax=stress.series.index.max(),
        )

        rfunc1.update_rfunc_settings(up=up)
        self.rfunc1 = rfunc1

        rfunc2.update_rfunc_settings(up=up)
        self.rfunc2 = rfunc2
        self.tchange = Timestamp(tchange)

        self.freq = stress.settings["freq"]
        self.stress.append(stress)
        self.set_init_parameters()

    def set_init_parameters(self) -> None:
        """Internal method to set the initial parameters."""
        self.parameters = concat(
            [
                self.rfunc1.get_init_parameters("{}_1".format(self.name)),
                self.rfunc2.get_init_parameters("{}_2".format(self.name)),
            ],
            axis=0,
        )

        tmin = Timestamp.min.toordinal()
        tmax = Timestamp.max.toordinal()
        tchange = self.tchange.toordinal()

        self.parameters.loc[self.name + "_beta"] = (
            0.0,
            -np.inf,
            np.inf,
            True,
            self.name,
            "uniform",
        )
        self.parameters.loc[self.name + "_tchange"] = (
            tchange,
            tmin,
            tmax,
            False,
            self.name,
            "uniform",
        )
        self.parameters.name = self.name

    def simulate(
        self,
        p: ArrayLike,
        tmin: TimestampType | None = None,
        tmax: TimestampType | None = None,
        freq: str | None = None,
        dt: float = 1.0,
    ) -> Series:
        self.update_stress(tmin=tmin, tmax=tmax, freq=freq)
        rfunc1 = self.rfunc1.block(p[: self.rfunc1.nparam])
        rfunc2 = self.rfunc2.block(
            p[self.rfunc1.nparam : self.rfunc1.nparam + self.rfunc2.nparam]
        )

        stress = self.stress[0].series
        npoints = stress.index.size
        t = np.linspace(0, 1, npoints)
        beta = p[-2]

        sigma = stress.index.get_loc(Timestamp.fromordinal(int(p[-1]))) / npoints
        omega = 1 / (np.exp(beta * (t - sigma)) + 1)

        h1 = Series(
            data=fftconvolve(stress, rfunc1, "full")[:npoints],
            index=stress.index,
            name=f"{self.name}_1",
        )
        h2 = Series(
            data=fftconvolve(stress, rfunc2, "full")[:npoints],
            index=stress.index,
            name=f"{self.name}_2",
        )
        h = (omega * h1 + (1 - omega) * h2).rename(self.name)

        return h

    def to_dict(self, series: bool = True):
        """Method to export the ChangeModel object.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the object.

        Notes
        -----
        Settings and metadata are exported with the stress.
        """
        data = {
            "stress": self.stress[0].to_dict(series=series),
            "rfunc1": self.rfunc1.to_dict(),
            "rfunc2": self.rfunc2.to_dict(),
            "name": self.name,
            "tchange": self.tchange,
            "up": self.rfunc1.up,
        }
        return data



"""Tests for the timer module."""

import time
from collections.abc import Callable
from typing import Any
from unittest.mock import patch

import pytest

from pastas.timer import ExceededMaxSolveTime, SolveTimer


class TestSolveTimer:
    """Test class for SolveTimer."""

    def test_init(self) -> None:
        """Test initialization of SolveTimer."""
        # Test default initialization
        timer = SolveTimer()
        assert timer.max_time is None
        assert timer.desc == "Optimization progress"
        assert timer.total is None

        # Test custom initialization
        timer = SolveTimer(max_time=60, desc="Custom progress", total=100)
        assert timer.max_time == 60
        assert timer.desc == "Custom progress"
        assert timer.total == 100

    @patch("pastas.timer.tqdm.update")
    def test_timer_callback(self, mock_update: Any) -> None:
        """Test the timer callback function."""
        # Setup mock return value for parent update method
        mock_update.return_value = True

        # Create SolveTimer with no max_time
        timer = SolveTimer()

        # Call timer callback and verify update was called
        result = timer.timer(None)
        mock_update.assert_called_once_with(1)
        assert result is True

        # Reset mock and test with custom n value
        mock_update.reset_mock()
        timer.timer(None, n=5)
        mock_update.assert_called_once_with(5)

    @patch("pastas.timer.tqdm.__init__")
    def test_custom_kwargs_passed_to_parent(self, mock_init: Any) -> None:
        """Test that custom kwargs are passed to parent class."""
        mock_init.return_value = None

        # Initialize with custom kwargs
        SolveTimer(max_time=30, leave=False, position=1, colour="green")

        # Check if kwargs were passed to parent
        _, kwargs = mock_init.call_args
        assert kwargs.get("leave") is False
        assert kwargs.get("position") == 1
        assert kwargs.get("colour") == "green"


@pytest.mark.skipif(
    "tqdm" not in pytest.importorskip("tqdm").__file__,
    reason="Test requires tqdm to be installed directly, not as a vendored package",
)
def test_real_usage() -> None:
    """Test SolveTimer in a scenario closer to real usage."""

    # Create a mock function to simulate model.solve
    def mock_solve(callback: Callable | None = None) -> None:
        for i in range(5):
            if callback:
                callback(None)
            time.sleep(0.01)

    # Test with a small max_time that won't be exceeded
    with SolveTimer(max_time=1) as timer:
        mock_solve(callback=timer.timer)

    # Test with a max_time that will be exceeded
    with pytest.raises(ExceededMaxSolveTime):
        with SolveTimer(max_time=0.02) as timer:
            mock_solve(callback=timer.timer)



"""Tests for input validation across Pastas."""

import numpy as np
import pandas as pd
import pytest

import pastas as ps
from pastas.timeseries import TimeSeries


class TestInputValidation:
    """Test input validation for various Pastas components."""

    def test_invalid_time_series_inputs(self) -> None:
        """Test handling of invalid time series inputs."""
        # Test with empty series
        empty_series = pd.Series([], dtype=float)
        with pytest.raises(ValueError):
            TimeSeries(empty_series)

        # Test with series containing only NaNs
        nan_series = pd.Series([np.nan, np.nan], index=pd.date_range("2000", periods=2))
        with pytest.raises(ValueError):
            TimeSeries(nan_series)

    def test_invalid_model_parameters(self, ml_noisemodel: ps.Model) -> None:
        """Test handling of invalid model parameters."""
        # Use the ml_solved fixture from conftest.py
        ml = ml_noisemodel

        # Test invalid parameter name
        with pytest.raises(KeyError):
            ml.set_parameter("nonexistent_parameter", initial=1.0)

        # Test inconsistent bounds
        param_name = "constant_d"
        with pytest.raises(ValueError):
            # Setting pmin > pmax should raise error
            ml.set_parameter(param_name, initial=7.5, pmin=10.0, pmax=5.0)

        # Test invalid parameter value type
        with pytest.raises(TypeError):
            ml.set_parameter(param_name, initial="not_a_number")

        # Test setting invalid bounds types
        with pytest.raises(TypeError):
            ml.set_parameter(param_name, initial=5.0, pmin="invalid")

        with pytest.raises(TypeError):
            ml.set_parameter(param_name, initial=5.0, pmax="invalid")

        # Test that initial value respects bounds
        with pytest.raises(ValueError):
            ml.set_parameter(param_name, initial=0.0, pmin=1.0, pmax=5.0)

        with pytest.raises(ValueError):
            ml.set_parameter(param_name, initial=10.0, pmin=1.0, pmax=5.0)

    def test_invalid_solve_parameters(self) -> None:
        """Test handling of invalid solve parameters."""
        dates = pd.date_range("2000", "2001", freq="D")
        head = pd.Series(np.random.normal(0, 1, len(dates)), index=dates)
        ml = ps.Model(head)

        # Test invalid tmin/tmax
        with pytest.raises(ValueError):
            ml.solve(tmin="2002", tmax="2001")

        # Test invalid frequency
        with pytest.raises(ValueError):
            ml.solve(freq="invalid_freq")

    def test_incompatible_time_ranges(self) -> None:
        """Test handling of incompatible time ranges between series."""
        # Create series with non-overlapping time ranges
        dates1 = pd.date_range("2000", "2001", freq="D")
        dates2 = pd.date_range("2002", "2003", freq="D")

        head = pd.Series(np.random.normal(0, 1, len(dates1)), index=dates1)
        stress = pd.Series(np.random.normal(0, 1, len(dates2)), index=dates2)

        ml = ps.Model(head)
        sm = ps.StressModel(stress, rfunc=ps.Exponential(), name="stress")

        # Adding the stress model should work but produce a warning about no overlap
        ml.add_stressmodel(sm)

        # Solving should raise an error about time series extension
        with pytest.raises(ValueError, match="cannot be extended into past"):
            ml.solve()



"""The following methods may be used for the diagnostic checking of the residual time series of a calibrated (Pastas) model."""

from logging import getLogger

from numpy import arange, cumsum, finfo, median, nan, sqrt, zeros
from pandas import DataFrame, Series, date_range, infer_freq
from scipy.stats import chi2, norm, normaltest, shapiro

from pastas.stats.core import acf as get_acf
from pastas.timeseries_utils import _get_time_offset, get_equidistant_series_nearest

logger = getLogger(__name__)
__all__ = [
    "durbin_watson",
    "ljung_box",
    "runs_test",
    "stoffer_toloi",
    "diagnostics",
]


def durbin_watson(series: Series) -> float:
    """Durbin-Watson test for autocorrelation.

    Parameters
    ----------
    series: pandas.Series, optional
        residuals series.

    Returns
    -------
    dw_stat: float
        The method returns the Durbin-Watson test statistic.

    Notes
    -----
    The Durban Watson statistic ([durbin_1951]_, [Fahidy_2004]_) tests the
    null-hypothesis that the correlation between the noise values at lag one equals
    zero. The formula to calculate the Durbin-Watson statistic (DW) is:

    .. math::
        DW = \\frac{\\sum_{t=2}^{n}(\\upsilon_t-\\upsilon_{t-1}^2)}
        {\\sum_{t=1}^{n}\\upsilon_t^2}

    where $n$ is the number of values in the noise series. The test-statistic has a
    range :math:`0 \\geq DW \\leq 4`, where values of $DW < 2$ indicate a positive
    correlation and values of $DW > 2$ indicates negative autocorrelation. The
    Durbin-Watson test requires a constant time interval of the noise series and
    tests for autocorrelation at a lag of 1 time step.

    **Considerations for this test:**

    - The time series should have equidistant time steps.
    - The Durbin-Watson test tests for autocorrelation at lag 1 but not for
      larger time lags.
    - The test statistic for this test is difficult to compute and is usually
      obtained from pre-calculated tables.

    References
    ----------
    .. [durbin_1951] Durbin, J., & Watson, G. S. (1951). Testing for serial
       correlation in least squares regression. II. Biometrika, 38(1/2), 159-177.

    .. [Fahidy_2004] Fahidy, T. Z. (2004). On the Application of Durbin-Watson
      Statistics to Time-Series-Based Regression Models. CHEMICAL ENGINEERING
      EDUCATION, 38(1), 22-25.

    Examples
    --------
    >>> data = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                data=np.random.rand(1000))
    >>> result = ps.stats.durbin_watson(data)
    """
    if not infer_freq(series.index):
        logger.warning(
            "The Durbin-Watson test should only be used for time series with "
            "equidistant time steps."
        )

    rho = series.autocorr(lag=1)  # Take the first value of the ACF

    dw_stat = 2 * (1 - rho)
    p = nan  # NotImplementedYet
    return dw_stat, p


def ljung_box(
    series: Series, lags: int = 15, nparam: int = 0, full_output: bool = False
) -> tuple[float, float]:
    """Ljung-box test for autocorrelation.

    Parameters
    ----------
    series: pandas.Series, optional
        series to calculate the autocorrelation for that is used in the Ljung-Box test.
    lags: int, optional
        The maximum lag to compute the Ljung-Box test statistic for.
    nparam: int, optional
        Number of calibrated parameters in the model.
    full_output: bool, optional
        Return the result of the test as a boolean (True) or not (False).

    Returns
    -------
    q_stat: float
        The computed Q test statistic.
    pval: float
        The probability of the computed Q test statistic.

    Notes
    -----
    The Ljung-Box test [Ljung_1978]_ tests the null-hypothesis that a time series are
    independently distributed up to a desired time lag $k$ and is computed as follows:

    .. math::
        Q(k) = n (n + 2) \\sum_{k=1}^{h} \\frac{\\rho^2(k)}{n - k}

    where :math:`\\rho_k` is the autocorrelation at lag $k$, $h$ is the maximum lag
    used for calculation, and $n$ is the number of values in the noise series. The
    computed $Q$-statistic is then compared to a critical value computed from a
    :math:`\\chi^2_{\\alpha, h-p}` distribution with a significance level
    :math:`\\alpha` and $h-p$ degrees of freedom, where $h$ is the number of lags and
    $p$ the number of the noise model parameters.

    **Considerations for this test:**

    - The time series should have equidistant time steps. An adapted version
      of the Ljung-Box test is available through ps.stats.stoffer_toloi.
    - A potential problem of the Ljung-Box test is the low power of the test when
      testing for a large number of lags using a small sample size $n$. It has been
      suggested that suggested that :math:`k \\leq n/4` but also as low as :math:`k
      \\leq n/20`. If we are using daily groundwater levels observations, and we want
      to test for autocorrelation for lags up to one year (365 days) this means that
      we need between 4 and ten years of data.

    References
    ----------
    .. [Ljung_1978] Ljung, G. and Box, G. (1978). On a Measure of Lack of Fit in Time
       Series Models, Biometrika, 65, 297-303.

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> stat, p = ps.stats.ljung_box(res, lags=15)
    >>> if p > alpha:
    >>>    print("Failed to reject the Null-hypothesis, no significant"
    >>>          "autocorrelation. p =", p.round(2))
    >>> else:
    >>>    print("Reject the Null-hypothesis. p =", p.round(2))

    See Also
    --------
    pastas.stats.acf
        This method is called to compute the autocorrelation function.
    pastas.stats.stoffer_toloi
        Similar method but adapted for time series with missing data.
    """
    if not infer_freq(series.index):
        logger.warning(
            "Caution: The Ljung-Box test should only be used for time series with "
            "equidistant time steps. Consider using ps.stats.stoffer_toloi instead."
        )

    acf = get_acf(series, lags=lags, bin_method="regular")
    nobs = series.index.size

    # Drop zero-lag from the acf and drop nan-values as k > 0
    acf = acf.drop(0, errors="ignore").dropna()
    lags = arange(1, len(acf) + 1)

    q_stat = nobs * (nobs + 2) * cumsum(acf.values**2 / (nobs - lags))
    dof = max(lags[-1] - nparam, 1)
    pval = chi2.sf(q_stat, df=dof)

    if full_output:
        result = DataFrame(data={"Q Stat": q_stat, "P-value": pval}, index=acf.index)
        return result
    else:
        return q_stat[-1], pval[-1]


def runs_test(series: Series, cutoff: str = "median") -> tuple[float, float]:
    """Runs test for autocorrelation.

    Parameters
    ----------
    series: pandas.Series
        Time series to test for autocorrelation.
    cutoff: str or float, optional
        String set to "mean", "median", or a float value to use as the cutoff.

    Returns
    -------
    z_stat: float
        Runs test statistic.
    pval: float
        p-value for the test statistic, based on a normal distribution.

    Notes
    -----
    Wald and Wolfowitz developed [wald_1943]_ developed a distribution free test (
    i.e., no normal distribution is assumed) to test for autocorrelation. This test
    is also appropriate for non-equidistant time steps in the residuals time series.
    The Null-hypothesis is that the residual time series is a random sequence of
    positive and negative values. The alternative hypothesis is that they are
    non-random. The test statistic is computed as follows:

    .. math::
        Z = \\frac{R-\\bar{R}}{\\sigma_R}

    where $R$ is the number of runs, :math:`\\bar{R}` the expected number of runs and
    :math:`\\sigma_R` the standard deviation of the number of runs. A run is defined
    as the number of sequences of exclusively positive and negative values in the
    time series.

    **Considerations for this test:**

    - Test is also applicable to time series with non-equidistant time steps.

    References
    ----------
    .. [wald_1943] Wald, A., & Wolfowitz, J. (1943). An exact test for randomness in
       the non-parametric case based on serial correlation. The Annals of
       Mathematical Statistics, 14(4), 378-388.

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> stat, pval = ps.stats.runs_test(res)
    >>> if p > alpha:
    >>>     print("Failed to reject the Null-hypothesis, no significant"
    >>>           "autocorrelation. p =", p.round(2))
    >>> else:
    >>>     print("Reject the Null-hypothesis")
    """
    # Make dichotomous sequence
    r = series.copy().to_numpy()
    if cutoff == "mean":
        cutoff = r.mean()
    elif cutoff == "median":
        cutoff = median(r)
    elif isinstance(cutoff, float):
        pass
    else:
        msg = """Cutoff criterion %s is not implemented. Cutoff should be 'mean',
                 'median', or a float value."""
        logger.error(msg, cutoff)
        raise NotImplementedError(msg % cutoff)

    r[r > cutoff] = 1
    r[r < cutoff] = 0

    # Calculate number of positive and negative noise
    n_pos = r.sum()
    n_neg = r.size - n_pos

    # Calculate the number of runs
    runs = r[1:] - r[0:-1]
    n_runs = sum(abs(runs)) + 1

    # Calculate the expected number of runs and the standard deviation
    n_neg_pos = 2.0 * n_neg * n_pos
    n_runs_exp = n_neg_pos / (n_neg + n_pos) + 1
    n_runs_std = (n_neg_pos * (n_neg_pos - n_neg - n_pos)) / (
        (n_neg + n_pos) ** 2 * (n_neg + n_pos - 1)
    )

    # Calculate Z-statistic and pvalue
    z_stat = (n_runs - n_runs_exp) / sqrt(n_runs_std)
    pval = 2 * norm.sf(abs(z_stat))

    return z_stat, pval


def stoffer_toloi(
    series: Series,
    lags: int = 15,
    nparam: int = 0,
    freq: str = "D",
    snap_to_equidistant_timestamps: bool = False,
) -> tuple[float, float]:
    """Adapted Ljung-Box test to deal with missing data [stoffer_1992]_.

    Parameters
    ----------
    series: pandas.Series
        Time series to compute the adapted Ljung-Box statistic for.
    lags: int, optional
        the number of lags to compute the statistic for. Only lags for which a
        correlation is computed are used.
    nparam: int, optional
        Number of parameters of the noisemodel.
    freq: str, optional
        String with the frequency to resample the time series to.
    snap_to_equidistant_timestamps : bool, optional
        if False (default), a sample is taken from series with equidistant timesteps
        using pandas' reindex. Only values are kept that lie on those equidistant
        timestamps. If True, an equidistant time series is created taking as many
        values as possible from the original series which are then snapped to the
        nearest equidistant timestamp.

    Returns
    -------
    qm: float
        Adapted Ljung-Box test statistic.
    pval: float
        p-value for the test statistic, based on a chi-squared distribution.

    Notes
    -----
    Stoffer and Toloi [stoffer_1992]_ extended the Ljung-Box test to also work with
    missing data. The test statistic is computed as follows:

    .. math ::
        Q_k = n^2 \\sum_{k=1}^{h} \\frac{\\hat{\\rho}_k^2}{n-k}

    where :math:`\\hat{\\rho}_k` is the autocorrelation for lag $k$. When the
    residual time series have non-equidistant time steps it is recommended to use
    this test over the original Ljung-Box test.

    The Stoffer-Toloi test is strictly an adapted version of the Ljung-Box test to
    deal with missing data in a time series and not a time series with
    non-equidistant time steps. This means that the time series is updated to an
    equidistant time series by filling nan-values.

    **Considerations for this test:**

    - Test is also applicable to irregular time series.
    - The time step has to be chosen (e.g., Days). This should not be smaller than
      the smallest time step or the test will most likely fail to reject $H_0$ anyway.

    References
    ----------
    .. [stoffer_1992] Stoffer, D. S., & Toloi, C. M. (1992). A note on the
       Ljung—Box—Pierce stoffer_toloi statistic with missing data. Statistics &
       probability letters, 13(5), 391-396.

    Examples
    --------
    >>> data= pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> stat, p = ps.stats.stoffer_toloi(noise, lags=15, freq="D")
    >>> if p > alpha:
    >>>    print("Failed to reject the Null-hypothesis, no significant"
    >>>          "autocorrelation. p =", p.round(2))
    >>> else:
    >>>    print("Reject the Null-hypothesis")

    See Also
    --------
    pastas.timeseries_utils.get_equidistant_series_nearest
    """
    if snap_to_equidistant_timestamps:
        # create equidistant time series snapping values from the original series to
        # the nearest equidistant timestamp. No values are duplicated and data loss
        # is minimized.
        s = get_equidistant_series_nearest(series, freq, minimize_data_loss=True)
    else:
        # get equidistant sample from original time series, checks which time offset
        # is the most common to maximize the number of values taken from the original
        # series.
        t_offset = _get_time_offset(series.index, freq).value_counts().idxmax()
        new_idx = date_range(
            series.index[0].floor(freq) + t_offset,
            series.index[-1].floor(freq) + t_offset,
            freq=freq,
        )
        s = series.reindex(new_idx)
        # warn if more than 10% of data is lost in sample
        if s.dropna().index.size < (0.9 * series.dropna().index.size):
            msg = (
                "While selecting equidistant values from series with `as_freq` more "
                "than 10 %% of values were dropped. Consider setting "
                "`make_equidistant` to True."
            )
            logger.warning(msg)

    nobs = s.size
    z = (s - s.mean()).fillna(0.0)
    y = z.to_numpy()
    yn = s.notna().to_numpy()

    dz0 = (y**2).sum() / nobs
    da0 = (yn**2).sum() / nobs
    de0 = dz0 / da0

    # initialize, compute all correlation up to one year.
    nlags = min(365, nobs - 1)
    dz = zeros(nlags)
    da = zeros(nlags)
    de = zeros(nlags)

    for i in range(0, nlags):
        hh = y[: -i - 1] * y[i + 1 :]
        dz[i] = hh.sum() / nobs
        hh = yn[: -i - 1] * yn[i + 1 :]
        da[i] = hh.sum() / (nobs - i - 1)
        if abs(da[i]) > finfo(float).eps:
            de[i] = dz[i] / da[i]

    re = de / de0

    # remove correlation where no observations are available (de = 0)
    da = da[re != 0][:lags]
    re = re[re != 0][:lags]
    k = arange(1, len(re) + 1)

    # Compute the Q-statistic
    qm = nobs**2 * sum(da * re**2 / (nobs - k))

    dof = max(len(k) - nparam, 1)
    pval = chi2.sf(qm, df=dof)

    return qm, pval


def diagnostics(
    series: Series,
    alpha: float = 0.05,
    nparam: int = 0,
    lags: int = 15,
    stats: tuple = (),
    float_fmt: str = "{0:.2f}",
) -> DataFrame:
    """Methods to compute various diagnostics checks for a time series.

    Parameters
    ----------
    series: pandas.Series
        Time series to compute the diagnostics for.
    alpha: float, optional
        significance level to use for the hypothesis testing.
    nparam: int, optional
        Number of parameters of the noisemodel.
    lags: int, optional
        Maximum number of lags (in days) to compute the autocorrelation tests for.
    stats: tuple, optional
        Tuple with the diagnostic checks to perform. Not implemented yet.
    float_fmt: str
        String to use for formatting the floats in the returned DataFrame.

    Returns
    -------
    df: Pandas.DataFrame
        DataFrame with the information for the diagnostics checks. The final column
        in this DataFrame report if the Null-Hypothesis is rejected. If H0 is not
        rejected (=False) the data is in agreement with one of the properties of
        white noise (e.g., normally distributed).

    Notes
    -----
    Different tests are computed depending on the regularity of the time step of the
    provided time series. pd.infer_freq is used to determined if the time steps are
    regular.

    Examples
    --------
    >>> data = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> ps.stats.diagnostics(data)
    Out[0]:
                      Checks Statistic P-value  Reject H0
    Shapiroo       Normality      1.00    0.86      False
    D'Agostino     Normality      1.18    0.56      False
    Runs test      Autocorr.     -0.76    0.45      False
    Durbin-Watson  Autocorr.      2.02     nan      False
    Ljung-Box      Autocorr.      5.67    1.00      False

    In this example, the Null-hypothesis is not rejected and the data may be assumed
    to be white noise.
    """

    cols = ["Checks", "Statistic", "P-value"]
    df = DataFrame(index=stats, columns=cols)

    # Shapiroo-Wilk test for Normality
    stat, p = shapiro(series)
    df.loc["Shapiroo", cols] = "Normality", stat, p

    # D'Agostino test for Normality
    stat, p = normaltest(series)
    df.loc["D'Agostino", cols] = "Normality", stat, p

    # Runs test for autocorrelation
    stat, p = runs_test(series)
    df.loc["Runs test", cols] = "Autocorr.", stat, p

    # Do different tests depending on time step
    if infer_freq(series.index):
        # Ljung-Box test for autocorrelation
        stat, p = ljung_box(series, nparam=nparam, lags=lags)
        df.loc["Ljung-Box", cols] = "Autocorr.", stat, p

        # Durbin-Watson test for autocorrelation
        stat, p = durbin_watson(series)
        df.loc["Durbin-Watson", cols] = "Autocorr.", stat, p
    else:
        # Stoffer-Toloi for autocorrelation
        stat, p = stoffer_toloi(series, nparam=nparam, lags=lags)
        df.loc["Stoffer-Toloi", cols] = "Autocorr.", stat, p

    df["Reject H0 ($\\alpha$={:.2f})".format(alpha)] = df.loc[:, "P-value"] < alpha
    df.loc[:, "P-value"] = df.loc[:, "P-value"].apply(float_fmt.format)
    df.loc[:, "Statistic"] = df.loc[:, "Statistic"].apply(float_fmt.format)
    return df
